import unittest
import hashlib
from typing import List

from neo3 import vm, contracts, storage, blockchain, settings
from neo3.core.serialization import BinaryReader, BinaryWriter
from neo3.network import payloads
from neo3.contracts import syscall_name_to_int
from neo3.core import to_script_hash, types, cryptography, serialization, msgrouter
from tests.contracts.interop.utils import test_engine, test_block, test_tx, TestIVerifiable, contract_hash


class TestIVerifiable(payloads.IVerifiable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.script_hashes = [types.UInt160.zero()]

    def serialize(self, writer: BinaryWriter) -> None:
        pass

    def deserialize(self, reader: BinaryReader) -> None:
        pass

    def __len__(self):
        pass

    def serialize_unsigned(self, writer: serialization.BinaryWriter) -> None:
        pass

    def deserialize_unsigned(self, reader: serialization.BinaryReader) -> None:
        pass

    def get_script_hashes_for_verifying(self, snapshot: storage.Snapshot) -> List[types.UInt160]:
        return self.script_hashes


class RuntimeInteropTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_getplatform(self):
        engine = test_engine()
        engine.invoke_syscall_by_name("System.Runtime.Platform")
        item = engine.pop()
        self.assertIsInstance(item, vm.ByteStringStackItem)
        self.assertEqual(vm.ByteStringStackItem(b'NEO'), item)

    def test_gettrigger(self):
        engine = test_engine()
        engine.invoke_syscall_by_name("System.Runtime.GetTrigger")
        item = engine.pop()
        self.assertIsInstance(item, vm.IntegerStackItem)
        self.assertEqual(vm.IntegerStackItem(contracts.TriggerType.APPLICATION.value), item)

    def test_gettime(self):
        engine = test_engine(has_container=False, has_snapshot=True)
        b = test_block()
        engine.snapshot.persisting_block = b
        engine.invoke_syscall_by_name("System.Runtime.GetTime")
        item = engine.pop()
        self.assertIsInstance(item, vm.IntegerStackItem)
        self.assertEqual(vm.IntegerStackItem(b.timestamp), item)

    def test_getscriptcontainer(self):
        # first test against an invalid script container (IVerifiable, but not IOperable)
        engine = test_engine()
        container = payloads.Header._serializable_init()
        engine.script_container = container

        with self.assertRaises(ValueError) as context:
            engine.invoke_syscall_by_name("System.Runtime.GetScriptContainer")
        self.assertEqual("script container is not a valid IInteroperable type", str(context.exception))

        tx = test_tx(1)
        engine = test_engine()
        engine.script_container = tx

        sb = vm.ScriptBuilder()
        sb.emit_syscall(syscall_name_to_int("System.Runtime.GetScriptContainer"))
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()

        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack._items))
        item = engine.result_stack.pop()
        self.assertIsInstance(item, vm.ArrayStackItem)
        # we now have a Block that has been serialized, let's check the hash
        self.assertEqual(vm.ByteStringStackItem(tx.hash().to_array()), item[0])

    def test_getexecutingscripthash(self):
        engine = test_engine()
        sb = vm.ScriptBuilder()
        sb.emit(vm.OpCode.PUSH1)
        sb.emit(vm.OpCode.PUSH2)
        sb.emit(vm.OpCode.PUSH3)
        sb.emit_syscall(syscall_name_to_int("System.Runtime.GetExecutingScriptHash"))
        data = sb.to_array()

        engine.load_script(vm.Script(data))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(4, len(engine.result_stack._items))
        item = engine.result_stack.pop()
        self.assertEqual(to_script_hash(data).to_array(), item.to_array())

    def test_getcallingscripthash(self):
        """
        Testing this requires 2 contracts

        1) caller_contract: uses a System.Contract.Call to call callee_contract. This will set the calling script hash on the ExecutionContext of the callee_contract
        2) callee_contract: uses a System.Runtime.GetCallingScriptHash to return the calling script
        """
        sb = vm.ScriptBuilder()
        sb.emit_syscall(syscall_name_to_int("System.Runtime.GetCallingScriptHash"))
        callee_contract_script = sb.to_array()
        callee_nef = contracts.NEF(script=callee_contract_script)
        callee_contract_name = "callee_contract"
        callee_manifest = contracts.ContractManifest(callee_contract_name)
        callee_manifest.abi.methods = [
            contracts.ContractMethodDescriptor("test_func", 0, [], contracts.ContractParameterType.ANY, True)
        ]
        callee_contract_hash = contract_hash(types.UInt160.zero(), callee_nef.checksum, callee_contract_name)
        callee_contract = contracts.ContractState(1, callee_nef, callee_manifest, 0, callee_contract_hash)

        # create caller_contract script
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call(callee_contract.hash, "test_func")
        caller_script = sb.to_array()
        caller_nef = contracts.NEF(script=caller_script)
        caller_contract_name = "caller_contract"
        caller_manifest = contracts.ContractManifest(caller_contract_name)
        caller_contract_hash = contract_hash(types.UInt160.zero(), caller_nef.checksum, caller_contract_name)
        caller_contract = contracts.ContractState(2, caller_nef, caller_manifest, 0, caller_contract_hash)

        engine = test_engine(has_snapshot=True, default_script=False)
        engine.snapshot.contracts.put(callee_contract)
        engine.snapshot.contracts.put(caller_contract)
        engine.load_script(vm.Script(caller_script))
        engine.execute()

        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack))
        item = engine.result_stack.pop()
        self.assertEqual(to_script_hash(caller_nef.script).to_array(), item.to_array())

    def test_getentryscripthash(self):
        # entry script hash is set on a engine.load_script
        sb = vm.ScriptBuilder()
        sb.emit_syscall(syscall_name_to_int("System.Runtime.GetEntryScriptHash"))
        raw_script = sb.to_array()

        engine = test_engine(default_script=False)
        ctx = engine.load_script(vm.Script(raw_script))
        engine.execute()

        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack))
        item = engine.result_stack.pop()
        self.assertEqual(to_script_hash(raw_script).to_array(), item.to_array())

    def test_checkwitness_helper(self):
        engine = test_engine()
        tx = test_tx()
        engine.script_container = tx

        bad_hash = types.UInt160.zero()
        self.assertFalse(engine.checkwitness(bad_hash))

        # if scope is GLOBAL then checkwitness should always return true
        good_hash = tx.sender
        tx.signers[0].scope = payloads.WitnessScope.GLOBAL
        self.assertTrue(engine.checkwitness(good_hash))

    def test_checkwitness_helper_custom_contracts(self):
        engine = test_engine(has_snapshot=True, default_script=False)
        tx = test_tx()
        engine.script_container = tx

        sb = vm.ScriptBuilder()
        sb.emit(vm.OpCode.RET)
        caller_script = sb.to_array()
        engine.load_script(vm.Script(caller_script))

        # test for CUSTOM_CONTRACTS in scope
        tx.signers[0].scope = payloads.WitnessScope.CUSTOM_CONTRACTS
        tx.signers[0].allowed_contracts = [to_script_hash(caller_script)]

        self.assertTrue(engine.checkwitness(tx.sender))

        # now check it fails if the hash is not in the allowed_contracts_list
        tx.signers[0].allowed_contracts = [types.UInt160.zero()]
        self.assertFalse(engine.checkwitness(tx.sender))

    @unittest.SkipTest
    def test_checkwitness_custom_groups(self):
        """
        We need to setup 2 contracts
        1) caller_contract: uses a System.Contract.Call to call callee_contract. This will set the calling script hash on the ExecutionContext of the callee_contract
        2) callee_contract: uses a System.Runtime.CheckWitness
        """
        engine = test_engine(has_snapshot=True, default_script=False)
        tx = test_tx()
        engine.script_container = tx

        '''
        Create the callee contract, which does a checkwitness on its caller.
        Checkwitness shall only pass if the caller == "caller_contract" as defined by CustomGroup
        '''
        sb = vm.ScriptBuilder()
        sb.emit_syscall(syscall_name_to_int("System.Runtime.GetCallingScriptHash"))  # get script hash of "caller_contract"
        sb.emit_syscall(syscall_name_to_int("System.Runtime.CheckWitness"))
        callee_contract_script = sb.to_array()
        callee_nef = contracts.NEF(script=callee_contract_script)
        callee_contract_name = "callee_contract"
        callee_manifest = contracts.ContractManifest(callee_contract_name)
        callee_manifest.abi.methods = [
            contracts.ContractMethodDescriptor("test_func", 0, [], contracts.ContractParameterType.ANY, True)
        ]
        callee_contract_hash = contract_hash(tx.sender, callee_nef.checksum, callee_contract_name)
        callee_contract = contracts.ContractState(1, callee_nef, callee_manifest, 0, callee_contract_hash)

        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call(callee_contract.hash, "test_func")
        caller_script = sb.to_array()
        caller_nef = contracts.NEF(script=caller_script)
        caller_contract_name = "caller_contract"
        caller_manifest = contracts.ContractManifest(caller_contract_name)

        keypair = cryptography.KeyPair(private_key=b'\x01' * 32)
        signature = cryptography.sign(caller_script, keypair.private_key)
        caller_manifest.groups = [contracts.ContractGroup(
            public_key=keypair.public_key,
            signature=signature
        )]

        caller_contract_hash = contract_hash(tx.sender, caller_nef.checksum, caller_contract_name)
        caller_contract = contracts.ContractState(2, caller_nef, caller_manifest, 0, caller_contract_hash)
        engine.snapshot.contracts.put(caller_contract)
        engine.snapshot.contracts.put(callee_contract)
        engine.load_script(vm.Script(caller_script))

        tx.signers[0].scope = payloads.WitnessScope.CUSTOM_GROUPS
        tx.signers[0].allowed_groups = [keypair.public_key]

        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack))
        item = engine.result_stack.pop()
        self.assertTrue(item.to_boolean())

        # now try again but make sure it fails if the public key is not listed in the allowed groups
        engine = test_engine(has_snapshot=True, default_script=False)
        tx = test_tx()
        engine.script_container = tx

        engine.snapshot.contracts.put(caller_contract)
        engine.snapshot.contracts.put(callee_contract)
        engine.load_script(vm.Script(caller_script))

        tx.signers[0].scope = payloads.WitnessScope.CUSTOM_GROUPS
        keypair = cryptography.KeyPair(private_key=b'\x02' * 32)
        tx.signers[0].allowed_groups = [keypair.public_key]

        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack))
        item = engine.result_stack.pop()
        self.assertFalse(item.to_boolean())

    def test_checkwitness_called_by_entry(self):
        """
        We need to setup 2 contracts
        1) caller_contract: uses a System.Contract.Call to call callee_contract. This will set the calling script hash on the ExecutionContext of the callee_contract
        2) callee_contract: uses a System.Runtime.CheckWitness
        """
        engine = test_engine(has_snapshot=True, has_container=False, default_script=False)
        tx = test_tx()
        tx.signers[0].scope = payloads.WitnessScope.CALLED_BY_ENTRY
        engine.script_container = tx

        sb = vm.ScriptBuilder()
        sb.emit_push(tx.sender.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Runtime.CheckWitness"))
        callee_contract_script = sb.to_array()
        callee_nef = contracts.NEF(script=callee_contract_script)
        callee_manifest = contracts.ContractManifest()
        callee_manifest.abi.methods = [
            contracts.ContractMethodDescriptor("test_func", 0, [], contracts.ContractParameterType.ANY, True)
        ]
        callee_contract = contracts.ContractState(1, callee_nef, callee_manifest, 0, to_script_hash(callee_contract_script))

        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call(callee_contract.hash, "test_func")
        caller_script = sb.to_array()
        caller_nef = contracts.NEF(script=caller_script)
        caller_manifest = contracts.ContractManifest()
        caller_contract = contracts.ContractState(2, caller_nef, caller_manifest, 0, to_script_hash(caller_script))

        engine.snapshot.contracts.put(callee_contract)
        engine.snapshot.contracts.put(caller_contract)
        engine.load_script(vm.Script(caller_script))
        engine.execute()

        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack))
        item = engine.result_stack.pop()
        self.assertTrue(item.to_boolean())

    def test_checkwitness_helper_other_verifiable(self):
        engine = test_engine(has_snapshot=True, has_container=False, default_script=True)
        engine.script_container = TestIVerifiable()
        self.assertFalse(engine.checkwitness(types.UInt160(b'\x01' * 20)))

        # our test verifiable has 1 verifiable with a UInt160.zero() hash
        self.assertTrue(engine.checkwitness(types.UInt160.zero()))

    def test_checkwitness_with_public_key(self):
        kp = cryptography.KeyPair(b'\x01' * 32)
        redeemscript = contracts.Contract.create_signature_redeemscript(
            kp.public_key
        )
        intermediate_data = hashlib.sha256(redeemscript).digest()
        data = hashlib.new('ripemd160', intermediate_data).digest()
        hash_ = types.UInt160(data=data)

        engine = test_engine(has_snapshot=True, has_container=False, default_script=True)
        engine.script_container = TestIVerifiable()
        engine.script_container.script_hashes = [hash_]

        engine.push(vm.ByteStringStackItem(kp.public_key.to_array()))
        engine.invoke_syscall_by_name("System.Runtime.CheckWitness")

        self.assertEqual(1 , len(engine.current_context.evaluation_stack))
        item = engine.pop()
        self.assertTrue(item.to_boolean())

    def test_checkwitness_invalid_data(self):
        engine = test_engine()
        engine.push(vm.ByteStringStackItem(b''))
        with self.assertRaises(ValueError) as context:
            engine.invoke_syscall_by_name("System.Runtime.CheckWitness")
        self.assertEqual("Supplied CheckWitness data is not a valid hash", str(context.exception))

    def test_gasleft(self):
        engine = test_engine()
        engine.is_test_mode = True

        sb = vm.ScriptBuilder()
        sb.emit_syscall(syscall_name_to_int("System.Runtime.GasLeft"))
        data = sb.to_array()

        # test with test mode
        engine.load_script(vm.Script(data))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack._items))
        item = engine.result_stack.pop()
        self.assertEqual(vm.IntegerStackItem(-1), item)

        """
        using (var engine = ApplicationEngine.Create(TriggerType.Application, null, null, gas:500))
        using (var script = new ScriptBuilder())
        {
            script.EmitSysCall(ApplicationEngine.System_Runtime_GasLeft);
            engine.LoadScript(script.ToArray());
            Assert.AreEqual(VMState.HALT, engine.Execute());
            Assert.AreEqual(engine.ResultStack.Count, 1);
            Console.WriteLine(engine.ResultStack.Pop().GetInteger());
        }
        """
        # test with actual consumption
        engine = test_engine()
        engine.is_test_mode = False
        engine.gas_amount = 500
        # we can re-use the script
        engine.load_script(vm.Script(data))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack._items))
        item = engine.result_stack.pop()
        self.assertEqual(vm.IntegerStackItem(20), item)

    def test_get_invocation_counter_ok(self):
        """
        We need to setup 2 contracts
        1) caller_contract: uses a System.Contract.Call to call callee_contract. This will increase the invocation counter of the callee contract
        2) callee_contract: uses a System.Runtime.GetInvocationCounter
        """
        engine = test_engine(has_snapshot=True, has_container=False, default_script=False)

        sb = vm.ScriptBuilder()
        sb.emit_syscall(syscall_name_to_int("System.Runtime.GetInvocationCounter"))
        callee_contract_script = sb.to_array()
        callee_nef = contracts.NEF(script=callee_contract_script)
        callee_manifest = contracts.ContractManifest("contract1")
        callee_manifest.abi.methods = [
            contracts.ContractMethodDescriptor("test_func", 0, [], contracts.ContractParameterType.ANY, True)
        ]
        callee_contract = contracts.ContractState(1, callee_nef, callee_manifest, 0, contract_hash(types.UInt160.zero(), callee_nef.checksum, callee_manifest.name))

        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call(callee_contract.hash, "test_func")
        caller_script = sb.to_array()
        caller_nef = contracts.NEF(script=caller_script)
        caller_manifest = contracts.ContractManifest("contract2")
        caller_contract = contracts.ContractState(2, caller_nef, caller_manifest, 0, contract_hash(types.UInt160.zero(), caller_nef.checksum, caller_manifest.name))

        engine.snapshot.contracts.put(callee_contract)
        engine.snapshot.contracts.put(caller_contract)
        engine.load_script(vm.Script(caller_script))
        engine.execute()

        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack))
        item = engine.result_stack.pop()
        self.assertEqual(1, int(item))

    def test_runtime_log(self):
        message = ''

        def runtime_log(script_container: payloads.IVerifiable, msg: str):
            nonlocal message
            message = msg

        msgrouter.interop_log += runtime_log

        engine = test_engine()
        engine.push(vm.ByteStringStackItem('hello world'))
        engine.invoke_syscall_by_name("System.Runtime.Log")
        self.assertEqual('hello world', message)

        # try with too long message
        engine = test_engine()
        engine.push(vm.ByteStringStackItem(b'a' * (engine.MAX_NOTIFICATION_SIZE + 1)))
        with self.assertRaises(ValueError) as context:
            engine.invoke_syscall_by_name("System.Runtime.Log")
        self.assertEqual("Log message length (1025) exceeds maximum allowed (1024)", str(context.exception))

    def test_runtime_getnotifications(self):
        engine = test_engine()

        target_hash = types.UInt160.zero()
        state = vm.ArrayStackItem(engine.reference_counter)
        notif1 = (object(), target_hash, b'notif1', state)
        notif2 = (object(), types.UInt160(b'\x01' * 20), b'notif2', state)
        notif3 = (object(), target_hash, b'notif3', state)

        engine.notifications = [notif1, notif2, notif3]
        engine.push(vm.ByteStringStackItem(target_hash.to_array()))
        engine.invoke_syscall_by_name("System.Runtime.GetNotifications")

        self.assertEqual(1, len(engine.current_context.evaluation_stack))
        notification_items = engine.pop()
        item1 = notification_items[0]
        item2 = notification_items[1]
        self.assertEqual(item1[0].to_array(), types.UInt160.zero().to_array())
        self.assertEqual(item1[1].to_array(), b'notif1')
        self.assertEqual(item2[0].to_array(), types.UInt160.zero().to_array())
        self.assertEqual(item2[1].to_array(), b'notif3')

    def test_runtime_getnotifications_limit_exceeded(self):
        engine = test_engine()
        target_hash = types.UInt160.zero()
        state = vm.ArrayStackItem(engine.reference_counter)

        # we can't adjust the limit, so we need to exceed it
        for i in range(0, engine.MAX_STACK_SIZE + 1):
            engine.notifications.append((object(), target_hash, b'notif' + str(i).encode(), state))
        engine.push(vm.ByteStringStackItem(target_hash.to_array()))
        with self.assertRaises(ValueError) as context:
            engine.invoke_syscall_by_name("System.Runtime.GetNotifications")

        self.assertEqual("Notification count exceeds limits", str(context.exception))

    def test_runtime_notify(self):

        message = ''

        def runtime_notify(script_hash: types.UInt160, msg: str, state: vm.ArrayStackItem):
            nonlocal message
            message = msg

        msgrouter.interop_notify += runtime_notify

        engine = test_engine()
        engine.push(vm.ArrayStackItem(engine.reference_counter))  # state
        expected_message = 'my_notification_message'
        engine.push(vm.ByteStringStackItem(expected_message.encode()))  # event message
        engine.invoke_syscall_by_name("System.Runtime.Notify")

        self.assertEqual(message, expected_message)
        self.assertEqual(1, len(engine.notifications))
        item = engine.notifications[0]
        self.assertIsNone(item[0])  # == engine.script_container
        self.assertEqual(engine.current_scripthash, item[1])
        self.assertEqual(expected_message, item[2].decode())

    def test_runtime_notify_exceed_size(self):
        engine = test_engine()
        engine.push(vm.ArrayStackItem(engine.reference_counter))  # state
        engine.push(vm.ByteStringStackItem(b'\x01' * (engine.MAX_EVENT_SIZE + 1)))  # event messasge
        with self.assertRaises(ValueError) as context:
            engine.invoke_syscall_by_name("System.Runtime.Notify")
        self.assertEqual("Notify event name length (33) exceeds maximum allowed (32)", str(context.exception))

    def test_get_random_same_block(self):
        # tests taken from https://github.com/neo-project/neo/blob/2040537350c889562085505f104479a25a5ec5aa/tests/neo.UnitTests/SmartContract/UT_ApplicationEngine.Runtime.cs#L15
        settings.reset_settings_to_default()
        raw = bytes.fromhex("0000000000000000000000000000000000000000000000000001000000000000000000000000000000000000000001000112010000")
        tx = payloads.Transaction.deserialize_from_bytes(raw)

        # this little hack basically nullifies the singleton behaviour and ensures we create
        # a new instance every time we call it. This in turn gives us a clean backend/snapshot
        blockchain.Blockchain.__it__ = None
        snapshot = blockchain.Blockchain(store_genesis_block=True).currentSnapshot
        engine1 = contracts.ApplicationEngine(contracts.TriggerType.APPLICATION, tx, snapshot, 0, test_mode=True)
        engine1.load_script(vm.Script(b'\x01'))

        blockchain.Blockchain.__it__ = None
        snapshot = blockchain.Blockchain(store_genesis_block=True).currentSnapshot
        engine2 = contracts.ApplicationEngine(contracts.TriggerType.APPLICATION, tx, snapshot, 0, test_mode=True)
        engine2.load_script(vm.Script(b'\x01'))

        rand1 = engine1.invoke_syscall_by_name("System.Runtime.GetRandom")
        rand2 = engine1.invoke_syscall_by_name("System.Runtime.GetRandom")
        rand3 = engine1.invoke_syscall_by_name("System.Runtime.GetRandom")
        rand4 = engine1.invoke_syscall_by_name("System.Runtime.GetRandom")
        rand5 = engine1.invoke_syscall_by_name("System.Runtime.GetRandom")

        rand6 = engine2.invoke_syscall_by_name("System.Runtime.GetRandom")
        rand7 = engine2.invoke_syscall_by_name("System.Runtime.GetRandom")
        rand8 = engine2.invoke_syscall_by_name("System.Runtime.GetRandom")
        rand9 = engine2.invoke_syscall_by_name("System.Runtime.GetRandom")
        rand10 = engine2.invoke_syscall_by_name("System.Runtime.GetRandom")

        self.assertEqual(225932872514876835587448704843370203748, rand1)
        self.assertEqual(190129535548110356450238097068474508661, rand2)
        self.assertEqual(48930406787011198493485648810190184269, rand3)
        self.assertEqual(66199389469641263539889463157823839112, rand4)
        self.assertEqual(217172703763162599519098299724476526911, rand5)

        self.assertEqual(rand1, rand6)
        self.assertEqual(rand2, rand7)
        self.assertEqual(rand3, rand8)
        self.assertEqual(rand4, rand9)
        self.assertEqual(rand5, rand10)



from __future__ import annotations
from enum import IntEnum
from typing import List
from neo3.core import types
from neo3.core import Size as s, utils
from neo3.core import serialization
import abc
from .verification import IVerifiable


class InventoryType(IntEnum):
    TX = 0x2b
    BLOCK = 0x2c
    CONSENSUS = 0x2d
    EXTENSIBLE = 0x2e


class InventoryPayload(serialization.ISerializable):
    """
    A payload used to share inventory hashes.

    See also:
        - :ref:`getblocks <message-usage-getblocks>`
        - :ref:`getdata <message-usage-getdata>`
        - :ref:`mempool <message-usage-mempool>`
    """

    def __init__(self, type: InventoryType, hashes: List[types.UInt256]):
        """
        Create payload.

        Args:
            type: indicator to what type of object the the hashes of this payload relate to.
            hashes: hashes of "type" objects.
        """
        self.type = type
        self.hashes = hashes

    def __len__(self):
        """ Get the total size in bytes of the object. """
        return s.uint8 + utils.get_var_size(self.hashes)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint8(self.type)
        writer.write_var_int(len(self.hashes))
        for h in self.hashes:  # type: types.UInt256
            writer.write_bytes(h.to_array())

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.type = InventoryType(reader.read_uint8())
        self.hashes = reader.read_serializable_list(types.UInt256)

    @classmethod
    def _serializable_init(cls):
        return cls(InventoryType.BLOCK, [])


class IInventory(IVerifiable):
    @abc.abstractmethod
    def hash(self) -> types.UInt256:
        """"""

    @property
    @abc.abstractmethod
    def inventory_type(self) -> InventoryType:
        """"""



from __future__ import annotations
import base64
import binascii
import orjson as json  # type: ignore
import json as pystd_json
import re
from typing import List, Callable, Optional, Dict, Any
from neo3 import contracts, vm
from neo3.core import serialization, types, IJson, cryptography, utils
from neo3.core.serialization import BinaryReader, BinaryWriter


class ContractGroup(IJson):
    """
    Describes a set of mutually trusted contracts.

    See Also: ContractManifest.
    """
    def __init__(self, public_key: cryptography.ECPoint, signature: bytes):
        self.public_key = public_key
        self.signature = signature

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.public_key == other.public_key and self.signature == other.signature

    def is_valid(self, contract_hash: types.UInt160) -> bool:
        """
        Validate if the group has agreed on allowing the specific contract_hash.

        Args:
            contract_hash:
        """
        return cryptography.verify_signature(contract_hash.to_array(),
                                             self.signature,
                                             self.public_key.encode_point(False),
                                             cryptography.ECCCurve.SECP256R1)

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            'pubkey': str(self.public_key),
            'signature': base64.b64encode(self.signature).decode()
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractGroup:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the signature length is not 64.
        """
        pubkey = contracts.validate_type(json['pubkey'], str)
        c = cls(
            public_key=cryptography.ECPoint.deserialize_from_bytes(binascii.unhexlify(pubkey)),
            signature=base64.b64decode(contracts.validate_type(json['signature'], str).encode('utf8'))
        )
        if len(c.signature) != 64:
            raise ValueError("Format error - invalid signature length")
        return c

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        struct = vm.StructStackItem(reference_counter)
        struct.append(vm.ByteStringStackItem(self.public_key.to_array()))
        struct.append(vm.ByteStringStackItem(self.signature))
        return struct


class ContractPermission(IJson):
    """
    Describes a single set of outgoing call restrictions for a 'System.Contract.Call' SYSCALL.
    It describes what other smart contracts the executing contract is allowed to call and what exact methods on the
    other contract are allowed to be called. This is enforced during runtime.

    Example:
        Contract A (the executing contract) wants to call method "x" on Contract B. The runtime will query the manifest
        of Contract A and ask if this is allowed. The Manifest will search through its permissions (a list of
        ContractPermission objects) and ask if it "is_allowed(target_contract, target_method)".
    """
    def __init__(self, contract: contracts.ContractPermissionDescriptor, methods: WildcardContainer):
        self.contract = contract
        self.methods = methods

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.contract == other.contract and self.methods == other.methods

    @classmethod
    def default_permissions(cls) -> ContractPermission:
        """
        Construct a ContractPermission which allows any contract and any method to be called.
        """
        return cls(contracts.ContractPermissionDescriptor(),  # with no parameters equals to Wildcard
                   WildcardContainer.create_wildcard())

    def is_allowed(self, target_contract: contracts.ContractState, target_method: str) -> bool:
        """
        Return if it is allowed to call `target_method` on the target contract.

        Args:
            target_contract: the contract state of the contract to be called.
            target_method: the method of the contract to be called.
        """
        if self.contract.is_hash:
            if not self.contract.contract_hash == target_contract.hash:
                return False
        elif self.contract.is_group:
            results = list(map(lambda p: p.public_key != self.contract.group, target_contract.manifest.groups))
            if all(results):
                return False
        return self.methods.is_wildcard or target_method in self.methods

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = self.contract.to_json()
        # because NEO C# returns a string from "method" instead of sticking to a standard interface
        json.update({'methods': self.methods.to_json()['wildcard']})
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractPermission:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if a method is zero length.
        """
        cpd = contracts.ContractPermissionDescriptor.from_json(json)
        json_wildcard = {'wildcard': json['methods']}
        methods = WildcardContainer.from_json(json_wildcard)
        for m in methods:
            if len(m) == 0:
                raise ValueError("Format error - methods cannot have length 0")
        return cls(cpd, methods)

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        struct = vm.StructStackItem(reference_counter)
        if self.contract.is_wildcard:
            struct.append(vm.NullStackItem())
        elif self.contract.is_hash:
            struct.append(vm.ByteStringStackItem(self.contract.contract_hash.to_array()))  # type: ignore
        else:
            struct.append(vm.ByteStringStackItem(self.contract.group.to_array()))  # type: ignore

        if self.methods.is_wildcard:
            struct.append(vm.NullStackItem())
        else:
            struct.append(
                vm.ArrayStackItem(reference_counter,
                                  list(map(lambda m: vm.ByteStringStackItem(m), self.methods)))  # type: ignore
            )
        return struct


class WildcardContainer(IJson):
    """
    An internal helper class for ContractManifest attributes.
    """
    def __init__(self, data: list = None):
        self._is_wildcard = False
        self._data = data if data else []

    def __contains__(self, item):
        return item in self._data

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self._data == other._data

    @classmethod
    def create_wildcard(cls) -> WildcardContainer:
        """
        Creates an instance that indicates any value is allowed.
        """
        instance = cls()
        instance._is_wildcard = True
        return instance

    @property
    def is_wildcard(self) -> bool:
        """
        Indicates if the container is configured to allow all values.
        """
        return self._is_wildcard

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        if self.is_wildcard:
            return {'wildcard': '*'}
        return {'wildcard': list(map(lambda d: str(d), self._data))}

    @classmethod
    def from_json(cls, json: dict):
        """
        Parse object out of JSON data.

        Note: if the value is not '*', and is a Python list, then it will assume
        that the list members are strings or convertible via str().

        If the wildcard should contain other data types, use the alternative `from_json_as_type()` method

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
            ValueError: if the data supplied cannot recreate a valid object.
        """
        value = json.get('wildcard', None)
        if value is None:
            raise ValueError(f"Invalid JSON - Cannot recreate wildcard from None")
        if value == '*':
            return WildcardContainer.create_wildcard()
        if isinstance(value, list):
            return WildcardContainer(data=list(map(lambda d: str(d), value)))
        raise ValueError(f"Invalid JSON - Cannot deduce WildcardContainer type from: {value}")

    @classmethod
    def from_json_as_type(cls, json: dict, conversion_func: Callable):
        """
            Parse object out of JSON data.

            Note: if the value is not '*', and is a Python list, then it will use `conversion_func` to
            parse the members into the expected types.

        Args:
            json: a dictionary.
            conversion_func: a callable that takes 1 argument, which is the element in the value list

            Example with UInt160:
                {'wildcard': ['0xa400ff00ff00ff00ff00ff00ff00ff00ff00ff01', 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa']}

                 the first call has as argument '0xa400ff00ff00ff00ff00ff00ff00ff00ff00ff01'.
                 to process this example call
                 WildcardContainer.from_json_as_type(json_data, lambda f: types.UInt160.from_string(f))

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
            ValueError: if the data supplied cannot recreate a valid object.
        """
        value = json.get('wildcard', None)
        if value is None:
            raise ValueError(f"Invalid JSON - Cannot recreate wildcard from None")
        if value == '*':
            return WildcardContainer.create_wildcard()
        if isinstance(value, list):
            return WildcardContainer(data=list(map(lambda d: conversion_func(d), value)))
        raise ValueError(f"Invalid JSON - Cannot deduce WildcardContainer type from: {value}")


class ContractManifest(serialization.ISerializable, IJson):
    """
    A description of a smart contract's abilities (callable methods & events) as well as a set of restrictions
    describing what external contracts and methods are allowed to be called.

    For more information see:
    https://github.com/neo-project/proposals/blob/3e492ad05d9de97abb6524fb9a73714e2cdc5461/nep-15.mediawiki
    """
    #: The maximum byte size after serialization to be considered valid a valid contract.
    MAX_LENGTH = 0xFFFF

    def __init__(self, name: Optional[str] = None):
        #: Contract name
        self.name: str = name if name else ""
        #: A group represents a set of mutually trusted contracts. A contract will trust and allow any contract in the
        #: same group to invoke it.
        self.groups: List[ContractGroup] = []

        #: The list of NEP standards supported e.g. "NEP-3"
        self.supported_standards: List[str] = []

        #: For technical details of ABI, please refer to NEP-14: NeoContract ABI.
        #: https://github.com/neo-project/proposals/blob/d1f4e9e1a67d22a5755c45595121f80b0971ea64/nep-14.mediawiki
        self.abi: contracts.ContractABI = contracts.ContractABI(
            events=[],
            methods=[]
        )

        #: Permissions describe what external contract(s) and what method(s) on these are allowed to be invoked.
        self.permissions: List[contracts.ContractPermission] = [contracts.ContractPermission.default_permissions()]

        # Update trusts/safe_methods with outcome of https://github.com/neo-project/neo/issues/1664
        # Unfortunately we have to add this nonsense logic or we get deviating VM results.
        self.trusts = WildcardContainer()  # for UInt160 types

        #: Optional user defined data
        self.extra: Optional[Dict] = None

    def __len__(self):
        return utils.get_var_size(str(self.to_json()).replace(' ', ''))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.name == other.name
                and self.groups == other.groups
                and self.abi == other.abi
                and self.permissions == other.permissions
                and self.trusts == other.trusts
                and self.extra == other.extra)

    def __str__(self):
        return json.dumps(self.to_json()).decode()

    def serialize(self, writer: BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_string(str(self))

    def deserialize(self, reader: BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self._deserialize_from_json(json.loads(reader.read_var_string(self.MAX_LENGTH)))

    def _deserialize_from_json(self, json: dict) -> None:
        if json['name'] is None:
            self.name = ""
        else:
            self.name = contracts.validate_type(json['name'], str)
        self.abi = contracts.ContractABI.from_json(json['abi'])
        self.groups = list(map(lambda g: ContractGroup.from_json(g), json['groups']))

        if len(json['features']) != 0:
            raise ValueError("Manifest features is reserved and cannot have any content at this time")

        self.supported_standards = list(map(lambda ss: contracts.validate_type(ss, str), json['supportedstandards']))
        self.permissions = list(map(lambda p: ContractPermission.from_json(p), json['permissions']))

        if json['trusts'] == '*':
            self.trusts = WildcardContainer.create_wildcard()
        else:
            self.trusts = WildcardContainer.from_json_as_type(
                {'wildcard': json['trusts']},
                lambda t: contracts.ContractPermissionDescriptor.from_json({'contract': t}))

        # converting json key/value back to default WildcardContainer format
        self.extra = json['extra']

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        if self.trusts.is_wildcard:
            trusts = '*'
        else:
            trusts = list(map(lambda m: m.to_json()['contract'], self.trusts))  # type: ignore
        json: Dict[str, Any] = {
            "name": self.name if self.name else None,
            "groups": list(map(lambda g: g.to_json(), self.groups)),
            "features": {},
            "supportedstandards": self.supported_standards,
            "abi": self.abi.to_json(),
            "permissions": list(map(lambda p: p.to_json(), self.permissions)),
            "trusts": trusts,
            "extra": self.extra
        }
        return json

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        struct = vm.StructStackItem(reference_counter)
        struct.append(vm.ByteStringStackItem(self.name))
        struct.append(vm.ArrayStackItem(reference_counter,
                                        list(map(lambda g: g.to_stack_item(reference_counter), self.groups)))
                      )
        struct.append(vm.MapStackItem(reference_counter))
        struct.append(vm.ArrayStackItem(reference_counter,
                                        list(map(lambda s: vm.ByteStringStackItem(s), self.supported_standards)))
                      )
        struct.append(self.abi.to_stack_item(reference_counter))
        struct.append(vm.ArrayStackItem(reference_counter,
                                        list(map(lambda p: p.to_stack_item(reference_counter), self.permissions)))
                      )
        if self.trusts.is_wildcard:
            struct.append(vm.NullStackItem())
        else:
            struct.append(
                vm.ArrayStackItem(reference_counter,
                                  list(map(lambda t: vm.ByteStringStackItem(t.to_array()),
                                           self.trusts)))  # type: ignore
            )
        if self.extra is None:
            struct.append(vm.ByteStringStackItem("null"))
        else:
            # find \uxxxx sequences
            p = re.compile(r'\\u[a-z0-9]{4}')
            s = pystd_json.dumps(self.extra, separators=(',', ':'))
            # uppercase the values such that the byte array matches C#
            fixed = p.sub(lambda match: f"\\u{match.group()[2:].upper()}", s)
            struct.append(vm.ByteStringStackItem(fixed))
        return struct

    @classmethod
    def from_json(cls, json: dict):
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raise:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the manifest name property has an incorrect format.
            ValueError: if the manifest support standards contains an string
        """
        manifest = cls()
        manifest._deserialize_from_json(json)
        if manifest.name is None or len(manifest.name) == 0:
            raise ValueError("Format error - invalid 'name'")
        for s in manifest.supported_standards:
            if len(s) == 0:
                raise ValueError("Format error - supported standards cannot be zero length")
        return manifest

    def is_valid(self, contract_hash: types.UInt160) -> bool:
        """
        Validates the if any in the manifest groups signed the requesting `contract_hash` as permissive.

        An example use-case is to allow creation and updating of smart contracts by a select group.

        Args:
            contract_hash:
        """
        result = list(map(lambda g: g.is_valid(contract_hash), self.groups))
        return all(result)

    def can_call(self, target_contract: contracts.ContractState, target_method: str) -> bool:
        results = list(map(lambda p: p.is_allowed(target_contract, target_method), self.permissions))
        return any(results)

    @classmethod
    def _serializable_init(cls):
        return cls()



from __future__ import annotations
import abc
from copy import deepcopy
from enum import Enum, auto
from typing import Optional, Iterator, Tuple, TypeVar, Any, List, TYPE_CHECKING
from neo3.core import types, serialization
from neo3 import storage
from contextlib import suppress

if TYPE_CHECKING:
    from neo3.network import payloads
    from neo3 import contracts

TKey = TypeVar('TKey', bound='serialization.ISerializable')
TValue = TypeVar('TValue', bound='serialization.ISerializable')


class TrackState(Enum):
    NONE = auto()
    ADDED = auto()
    CHANGED = auto()
    DELETED = auto()


class Trackable:
    def __init__(self, key: TKey, value: TValue, state: TrackState):
        self.key = key
        self.item = value
        self.state = state


class CachedAccess:
    def __init__(self, db):
        self._dictionary = {}
        self._changeset = set()
        self._db = db
        self._internal_get = None
        self._internal_try_get = None
        self._internal_all = None
        self._internal_seek = None

    def __getitem__(self, key):
        trackable = self._dictionary.get(key, None)  # type: Trackable
        if trackable is not None:
            if trackable.state == TrackState.DELETED:
                raise KeyError
        else:
            trackable = Trackable(key, self._internal_get(key), TrackState.NONE)
            self._dictionary.update({key: trackable})
        return trackable.item

    def _put(self, key, value):
        trackable = self._dictionary.get(key, None)  # type: Trackable
        if trackable is not None:
            if trackable.state != TrackState.DELETED:
                raise ValueError("Value already exists")
        else:
            if self._internal_try_get(key) is not None:
                raise ValueError("Value already exists")

        if trackable is None:
            trackable = Trackable(key, value, TrackState.ADDED)
        else:
            trackable = Trackable(key, value, TrackState.CHANGED)

        self._dictionary.update({key: trackable})
        self._changeset.add(key)

    def _get(self, key, read_only=False):
        """
        Return the value stored with `key`.

        Args:
            key: the identifier that can be used to retrieve the stored value
            read_only:

        Raises:
            KeyError if the key is not found.
        """
        value = self[key]
        if value is None:
            raise KeyError

        if read_only:
            # make sure that even trackables in ADDED/CHANGED state can't be modified
            return deepcopy(value)

        if self._dictionary[key].state == TrackState.NONE:
            self._dictionary[key].state = TrackState.CHANGED
            self._changeset.add(key)
        if self._dictionary[key].state == TrackState.DELETED:
            self._dictionary[key].state = TrackState.CHANGED
        return value

    def _delete(self, key) -> None:
        """
        Delete the value stored under `key`.

        Args:
            key: the identifier that can be used to identify the stored value
        """
        trackable = self._dictionary.get(key, None)  # type: Trackable
        if trackable is not None:
            if trackable.state == TrackState.ADDED:
                self._dictionary.pop(key)
                self._changeset.remove(key)
            else:
                trackable.state = TrackState.DELETED
                self._changeset.add(key)
        else:
            item = self._internal_try_get(key)
            if item is None:
                return

            self._dictionary.update({key: Trackable(key, item, TrackState.DELETED)})
            self._changeset.add(key)

    @abc.abstractmethod
    def commit(self):
        """ Persist changes"""

    @abc.abstractmethod
    def create_snapshot(self):
        """ Deep copy. """

    def get_changeset(self):
        """ Get the changes since creation or since last committing. """
        for key in self._changeset:
            yield self._dictionary[key]


class AttributeCache:
    def __init__(self):
        self._item = None
        self._state = TrackState.NONE

    def put(self, item) -> None:
        self._item = item

    def get(self, read_only=False) -> Any:
        if self._item is None:
            self._item = self._get_internal()

        if read_only:
            return deepcopy(self._item)
        else:
            self._state = TrackState.CHANGED
            return self._item

    def commit(self) -> None:
        if self._state == TrackState.CHANGED:
            self._update_internal(self._item)

    def create_snapshot(self) -> CloneAttributeCache:
        return CloneAttributeCache(self)

    @abc.abstractmethod
    def _get_internal(self):
        """ Return the value from the real backend. """

    @abc.abstractmethod
    def _update_internal(self, value):
        """ Update the value in the real backend. """


class CachedBlockAccess(CachedAccess):
    def __init__(self, db):
        super(CachedBlockAccess, self).__init__(db)
        self._height_hash_mapping = {}
        self._internal_get = self._db._internal_block_get
        self._internal_try_get = self._db._internal_block_try_get
        self._internal_all = self._db._internal_block_all
        self._internal_get_by_height = self._db._internal_block_get_by_height

    def put(self, block: payloads.Block) -> None:
        """
        Store a block.

        Args:
            block: instance.

        Raises:
            ValueError: if a duplicate item is found.
        """
        block_hash = block.hash()
        super(CachedBlockAccess, self)._put(block_hash, block)
        self._height_hash_mapping.update({block.index: block_hash})

    def get(self, hash: types.UInt256, read_only=False) -> payloads.Block:
        """
        Retrieve a block.

        Args:
            hash: block hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.

        Raises:
            KeyError: if the item is not found.
        """
        block = super(CachedBlockAccess, self)._get(hash, read_only)
        self._height_hash_mapping.update({block.index: block.hash()})
        return block

    def get_by_height(self, height: int, read_only=False) -> payloads.Block:
        """
        Retrieve a block by its height.

        Args:
            height: target block index/height.
            read_only: set to True to safeguard against return value modifications being persisted when committing.

        Raises:
            KeyError: if the item is not found.
        """
        block_hash = self._height_hash_mapping.get(height, None)
        if block_hash is None:
            block_hash = self._internal_get_by_height(height).hash()  # raises KeyError if block is not found
        return self.get(block_hash, read_only)

    def try_get(self, hash: types.UInt256, read_only=False) -> Optional[payloads.Block]:
        """
        Try to retrieve a block.

        Args:
            hash: block hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        try:
            return self.get(hash, read_only)
        except KeyError:
            return None

    def try_get_by_height(self, height: int, read_only=False) -> Optional[payloads.Block]:
        """
        Try to retrieve a block by its height.

        Args:
            height: block index.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        try:
            block = self.get_by_height(height, read_only)
        except KeyError:
            return None

        return self.try_get(block.hash(), read_only)

    def delete(self, hash: types.UInt256) -> None:
        """
        Remove a block.

        Args:
            hash: block hash.
        """
        super(CachedBlockAccess, self)._delete(hash)
        with suppress(KeyError):
            self._height_hash_mapping.pop(hash)

    def all(self) -> Iterator[payloads.Block]:
        """
        Retrieve all blocks (readonly)
        """
        blocks = []
        for block in self._internal_all():
            if block.hash() not in self._dictionary:
                blocks.append(block)

        for k, v in self._dictionary.items():
            if v.state != TrackState.DELETED:
                blocks.append(deepcopy(v.item))

        blocks.sort(key=lambda block: block.hash().to_array())
        for block in blocks:
            yield block
        return None


# It is currently unclear why these do not persist when they're class attributes to CachedContractAccess
# Keeping them here for the time being
_gas_token_contract_state = None
_neo_token_contract_state = None


class CachedContractAccess(CachedAccess):
    _gas_token_script_hash = types.UInt160.from_string("d2a4cff31913016155e38e474a2c06d08be276cf")
    _neo_token_script_hash = types.UInt160.from_string("ef4073a0f2b305a38ec4050e4d3d28bc40ea63f5")

    def __init__(self, db):
        super(CachedContractAccess, self).__init__(db)
        self._internal_get = self._db._internal_contract_get
        self._internal_try_get = self._db._internal_contract_try_get
        self._internal_all = self._db._internal_contract_all

    def put(self, contract: contracts.ContractState) -> None:
        """	
        Store a contract.	
        Args:	
            contract: contract state instance.
        Raises:	
            ValueError: if a duplicate item is found.	
        """
        super(CachedContractAccess, self)._put(contract.hash, contract)

    def get(self, hash: types.UInt160, read_only=False) -> contracts.ContractState:
        """
        Retrieve a contract.
        Args:
            hash: unique contract identifier.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        Raises:
            KeyError: if the item is not found.
        """
        global _gas_token_contract_state, _neo_token_contract_state
        if hash == self._gas_token_script_hash:
            if _gas_token_contract_state is None:
                _gas_token_contract_state = super(CachedContractAccess, self)._get(hash, read_only)
            return _gas_token_contract_state
        elif hash == self._neo_token_script_hash:
            if _neo_token_contract_state is None:
                _neo_token_contract_state = super(CachedContractAccess, self)._get(hash, read_only)
            return _neo_token_contract_state
        return super(CachedContractAccess, self)._get(hash, read_only)

    def try_get(self, hash: types.UInt160, read_only=False) -> Optional[contracts.ContractState]:
        """
        Try to retrieve a contract.
        Args:
            hash: unique contract identifier.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        try:
            return self.get(hash, read_only)
        except KeyError:
            return None

    def delete(self, hash: types.UInt160) -> None:
        """
        Remove a transaction.
        Args:
            hash: unique contract identifier.
        """
        super(CachedContractAccess, self)._delete(hash)

    def all(self) -> Iterator[contracts.ContractState]:
        """	
        Retrieve all contracts (readonly)	
        """
        contracts = []
        for contract in self._internal_all():
            if contract.hash not in self._dictionary:
                contracts.append(contract)

        for k, v in self._dictionary.items():
            if v.state != TrackState.DELETED:
                contracts.append(deepcopy(v.item))

        contracts.sort(key=lambda contract: contract.hash.to_array())
        for contract in contracts:
            yield contract
        return None


class _Enumerator:
    _default_pair = (storage.StorageKey(0, b''), storage.StorageItem(b''))

    def __init__(self, iter: Iterator):
        self.iter = iter
        self.value = None

    def move_next(self) -> bool:
        try:
            self.value = next(self.iter)
            return True
        except StopIteration:
            self.value = None
            return False

    def current(self) -> Tuple[storage.StorageKey, storage.StorageItem]:
        if self.value is None:
            return self._default_pair
        return self.value


class CachedStorageAccess(CachedAccess):
    def __init__(self, db):
        super(CachedStorageAccess, self).__init__(db)
        self._internal_get = self._db._internal_storage_get
        self._internal_try_get = self._db._internal_storage_try_get
        self._internal_all = self._db._internal_storage_all
        self._internal_find = self._db._internal_storage_find
        self._internal_seek = self._db._internal_storage_seek

    def put(self, key: storage.StorageKey, value: storage.StorageItem) -> None:
        """
        Store the value under the given key.

        Raises:
            ValueError: if the key already exists.

        Args:
            key: identifier to store the value under.
            value: the value to be persisted.
        """
        super(CachedStorageAccess, self)._put(key, value)

    def update(self, key: storage.StorageKey, value: storage.StorageItem) -> None:
        """
        Update the value under the given key.

        This is a convenience function which first deletes the old data and then puts the new data

        Args:
            key: identifier to store the value under.
            value: the value to be persisted.
        """
        item = self.try_get(key)
        if item:
            item.value = value.value
        else:
            super(CachedStorageAccess, self)._put(key, value)

    def get(self, key: storage.StorageKey, read_only=False) -> storage.StorageItem:
        """
        Retrieve a value from storage.

        Args:
            key: identifier.
            read_only: set to True to safeguard against return value modifications being persisted when committing.

        Raises:
            KeyError: if the item is not found.
        """
        return super(CachedStorageAccess, self)._get(key, read_only)

    def try_get(self, key: storage.StorageKey, read_only=False) -> Optional[storage.StorageItem]:
        """
        Try to retrieve a value from storage.

        Args:
            key: identifier.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        try:
            return self.get(key, read_only)
        except KeyError:
            return None

    def delete(self, key: storage.StorageKey) -> None:
        """
        Remove a key/value pair from storage

        Args:
            key: identifier to locate value.
        """
        super(CachedStorageAccess, self)._delete(key)

    def all(self, contract_id: Optional[int] = None) -> Iterator[Tuple[storage.StorageKey,
                                                                       storage.StorageItem]]:
        """
        Retrieve all storage key/value pairs, sorted by key (readonly).

        Note:
            Return values are sorted to give deterministic behaviour in smart contracts.

        Args:
            contract_id: smart contract id to limit results to. If not specified, returns for all
            contracts.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        pairs = []
        for k, v in self._internal_all(contract_id):
            if k not in self._dictionary:
                pairs.append((k, v))

        for k, v in self._dictionary.items():
            if v.state != TrackState.DELETED:
                pairs.append((deepcopy(v.key), deepcopy(v.item)))

        pairs.sort(key=lambda keypair: keypair[0].to_array())
        for pair in pairs:
            yield pair
        return None

    def find(self, key_prefix: bytes) -> Iterator[Tuple[storage.StorageKey, storage.StorageItem]]:
        """
        Retrieve all storage key/value pairs (readonly).

        Args:
            key_prefix: the prefix part of the storage.StorageKey.key to look for.

        """
        for key, value in self.seek(key_prefix):
            if key.to_array().startswith(key_prefix):
                yield key, value
            else:
                return None

    def find_range(self,
                   start: bytes,
                   end: bytes,
                   direction: str = "forward"
                   ) -> Iterator[Tuple[storage.StorageKey, storage.StorageItem]]:
        """
        Find all data within a given range.

        Args:
            start: the data prefix to start from
            end: the post fix to end the search results at
            direction: determines the database search order. Can be "forward" or "reverse"

        Returns:
            An iterator with the results

        See Also:
            * :func:`~neo3.storage.cache.CachedStorageAccess.seek`
        """
        comperator = storage.NEOByteCompare(direction)
        for key, value in self.seek(start, direction):
            if comperator.compare(key.to_array(), end) < 0:
                yield key, value
            else:
                return None

    def seek(self, key_prefix: bytes, direction="forward"
             ) -> Iterator[Tuple[storage.StorageKey, storage.StorageItem]]:
        """
        Find all data starting with a prefix.

        Args:
            key_prefix: the prefix the data must have
            direction: determines the database search order. Can be "forward" or "reverse"

        Returns:
            An iterator with the results

        See Also:
            * :func:`~neo3.storage.cache.CachedStorageAccess.find_range`

        """
        # always read only
        comperator = storage.NEOByteCompare(direction)

        cached: List[Tuple[storage.StorageKey, storage.StorageItem]] = []
        cached_keys: List[storage.StorageKey] = []
        for key, value in self._dictionary.items():
            if value.state != TrackState.DELETED and (
                    len(key_prefix) == 0 or comperator.compare(key.to_array(), key_prefix) >= 0):
                cached.append((key, value.item))
                cached_keys.append(key)

        if direction == "forward":
            cached_sorted = sorted(cached, key=lambda pair: pair[0].to_array())
        else:
            cached_sorted = sorted(cached, key=lambda pair: pair[0].to_array(), reverse=True)

        e1 = _Enumerator(iter(cached_sorted))
        e2 = _Enumerator(self._internal_seek(key_prefix, direction))

        c1 = e1.move_next()
        c2 = e2.move_next()

        i1 = e1.current()
        i2 = e2.current()

        comperator = storage.NEOByteCompare(direction)

        while c1 or c2:
            # filter out duplicates from internal if already found in cached
            if c2 and i2[0] in cached_keys:
                c2 = e2.move_next()
                i2 = e2.current()
            elif not c2 or (c1 and comperator.compare(i1[0].to_array(), i2[0].to_array()) < 0):
                # i1 values come from the cache by reference, make them "read-only"
                yield deepcopy(i1[0]), deepcopy(i1[1])
                c1 = e1.move_next()
                i1 = e1.current()
            else:
                # i2 values are new objects deserialized from storage (in the case of leveldb)
                yield i2[0], i2[1]
                c2 = e2.move_next()
                i2 = e2.current()


class CachedTXAccess(CachedAccess):
    def __init__(self, db):
        super(CachedTXAccess, self).__init__(db)
        self._internal_get = self._db._internal_transaction_get
        self._internal_try_get = self._db._internal_transaction_try_get
        self._internal_all = self._db._internal_transaction_all

    def put(self, tx: payloads.Transaction) -> None:
        """
        Store a transaction.

        Args:
            tx: instance.

        Raises:
            ValueError: if a duplicate item is found.
        """
        super(CachedTXAccess, self)._put(tx.hash(), tx)

    def get(self, hash: types.UInt256, read_only=False) -> payloads.Transaction:
        """
        Retrieve a transaction.

        Args:
            hash: transaction hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.

        Raises:
            KeyError: if the item is not found.
        """
        return super(CachedTXAccess, self)._get(hash, read_only)

    def try_get(self, hash: types.UInt256, read_only=False) -> Optional[payloads.Transaction]:
        """
        Try to retrieve a transaction.

        Args:
            hash: transaction hash.
            read_only: set to True to safeguard against return value modifications being persisted when committing.
        """
        try:
            return self.get(hash, read_only)
        except KeyError:
            return None

    def delete(self, hash: types.UInt256) -> None:
        """
        Remove a transaction.

        Args:
            hash: transaction hash.
        """
        super(CachedTXAccess, self)._delete(hash)

    def all(self) -> Iterator[payloads.Transaction]:
        """
        Retrieve all transactions (readonly)
        """
        transactions = []
        for tx in self._internal_all():
            if tx.hash() not in self._dictionary:
                transactions.append(tx)

        for k, v in self._dictionary.items():
            if v.state != TrackState.DELETED:
                transactions.append(deepcopy(v.item))

        transactions.sort(key=lambda tx: tx.hash().to_array())
        for tx in transactions:
            yield tx
        return None


"""
Yes there is near duplicate code for all the Clone_xxx_Cache classes. 

A solution with a generic CloneCache and multiple inheritance quickly became hard to follow in part because Python 
requires all parent classes to be designed for multiple inheritance (where our case needs different super args). The 
solution using **kwargs in combination with no manual controllable MRO becomes a mess. 

Revisit later when there's more time to think about the problem again. 
"""


class CloneBlockCache(CachedBlockAccess):
    def __init__(self, db, inner_cache: CachedBlockAccess):
        super(CloneBlockCache, self).__init__(db)
        self.inner_cache = inner_cache
        self._internal_get = self._inner_cache_get
        self._internal_try_get = self._inner_cache_try_get
        self._internal_all = self._inner_cache_all

    def _inner_cache_try_get(self, hash, read_only=True):
        try:
            return self._inner_cache_get(hash, read_only)
        except KeyError:
            return None

    def _inner_cache_get(self, hash, read_only=True):
        return self.inner_cache.get(hash, read_only)

    def _inner_cache_all(self):
        return self.inner_cache.all()

    def commit(self) -> None:
        """
        Persist changes to the parent snapshot.
        """
        keys_to_delete: List[types.UInt256] = []
        for trackable in self.get_changeset():  # trackable.item: payloads.Block
            if trackable.state == TrackState.ADDED:
                self.inner_cache.put(trackable.item)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == TrackState.CHANGED:
                # This one is kind of useless unless we augment Block with additional attributes (like done with
                # Transaction) and the IDBImplementation class calls are updated to serialize_special(). Otherwise, any
                # attribute changes will modify the hash of the Block causing it to not be found.
                item = self.inner_cache.try_get(trackable.item.hash(), read_only=False)
                if item:
                    item.from_replica(trackable.item)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == TrackState.DELETED:
                self.inner_cache.delete(trackable.item.hash())
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()


class CloneContractCache(CachedContractAccess):
    def __init__(self, db, inner_cache: CachedContractAccess):
        super(CloneContractCache, self).__init__(db)
        self.inner_cache = inner_cache
        self._internal_get = self._inner_cache_get
        self._internal_try_get = self._inner_cache_try_get
        self._internal_all = self._inner_cache_all

    def _inner_cache_try_get(self, hash, read_only=True):
        try:
            return self._inner_cache_get(hash, read_only)
        except KeyError:
            return None

    def _inner_cache_get(self, hash, read_only=True):
        return self.inner_cache.get(hash, read_only)

    def _inner_cache_all(self):
        return self.inner_cache.all()

    def commit(self) -> None:
        """	
        Persist changes to the parent snapshot.	
        """
        keys_to_delete: List[types.UInt160] = []
        for trackable in self.get_changeset():  # trackable.item: contracts.ContractState
            if trackable.state == TrackState.ADDED:
                self.inner_cache.put(trackable.item)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == TrackState.CHANGED:
                item = self.inner_cache.try_get(trackable.item.hash, read_only=False)
                if item:
                    item.from_replica(trackable.item)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == TrackState.DELETED:
                self.inner_cache.delete(trackable.item.hash)
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()


class CloneStorageCache(CachedStorageAccess):
    def __init__(self, db, inner_cache: CachedStorageAccess):
        super(CloneStorageCache, self).__init__(db)
        self.inner_cache = inner_cache
        self._internal_get = self._inner_cache_get
        self._internal_try_get = self._inner_cache_try_get
        self._internal_all = self._inner_cache_all
        self._internal_find = self._inner_cache_find

    def _inner_cache_try_get(self, hash, read_only=True):
        try:
            return self._inner_cache_get(hash, read_only)
        except KeyError:
            return None

    def _inner_cache_get(self, hash, read_only=True):
        return self.inner_cache.get(hash, read_only)

    def _inner_cache_all(self, contract_script_hash):
        return self.inner_cache.all(contract_script_hash)

    def _inner_cache_find(self, contract_script_hash, key_prefix):
        return self.inner_cache.find(contract_script_hash, key_prefix)

    def commit(self) -> None:
        """
        Persist changes to the parent snapshot.
        """
        keys_to_delete: List[storage.StorageKey] = []
        for trackable in self.get_changeset():
            if trackable.state == TrackState.ADDED:
                self.inner_cache.put(trackable.key, trackable.item.clone())
                trackable.state = storage.TrackState.NONE
            elif trackable.state == TrackState.CHANGED:
                item = self.inner_cache.try_get(trackable.key, read_only=False)
                if item:
                    item.from_replica(trackable.item)
                    trackable.state = storage.TrackState.NONE
            elif trackable.state == TrackState.DELETED:
                self.inner_cache.delete(trackable.key)
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()


class CloneTXCache(CachedTXAccess):
    def __init__(self, db, inner_cache: CachedTXAccess):
        super(CloneTXCache, self).__init__(db)
        self.inner_cache = inner_cache
        self._internal_get = self._inner_cache_get
        self._internal_try_get = self._inner_cache_try_get
        self._internal_all = self._inner_cache_all

    def _inner_cache_try_get(self, hash, read_only=True):
        try:
            return self._inner_cache_get(hash, read_only)
        except KeyError:
            return None

    def _inner_cache_get(self, hash, read_only=True):
        return self.inner_cache.get(hash, read_only)

    def _inner_cache_all(self):
        return self.inner_cache.all()

    def commit(self) -> None:
        """
        Persist changes to the parent snapshot.
        """
        keys_to_delete: List[types.UInt256] = []
        for trackable in self.get_changeset():  # trackable.item: payloads.Transaction
            if trackable.state == TrackState.ADDED:
                self.inner_cache.put(trackable.item)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == TrackState.CHANGED:
                # Note that any changes to TX attributes that are serialized changes the hash and thus we won't find it
                # anymore in the cache.
                #
                # This means the changes will be persisted in the DB with a new entry!
                # We can however augment the TX class with e.g. VMState and BlockIndex without affecting the hash.
                # For all other cases the behaviour of a `transaction.get(hash)` is equal to
                # `transaction.get(hash, readonly=True) as no data will be persisted.
                item = self.inner_cache.try_get(trackable.item.hash(), read_only=False)
                if item:
                    item.from_replica(trackable.item)
                trackable.state = storage.TrackState.NONE
            elif trackable.state == TrackState.DELETED:
                self.inner_cache.delete(trackable.item.hash())
                keys_to_delete.append(trackable.key)
        for key in keys_to_delete:
            with suppress(KeyError):
                self._dictionary.pop(key)
        self._changeset.clear()


class CloneAttributeCache(AttributeCache):
    def __init__(self, inner_cache: AttributeCache):
        super(CloneAttributeCache, self).__init__()
        self._inner_cache = inner_cache

    def _get_internal(self):
        return self._inner_cache.get()

    def _update_internal(self, value):
        self._inner_cache.put(value)



from __future__ import annotations
import json
import os.path
from typing import List, Optional
from neo3.wallet.account import Account
from neo3.wallet.scrypt_parameters import ScryptParameters
from neo3.wallet.wallet import Wallet


class NEP6DiskWallet(Wallet):
    """
    A specialised wallet for persisting wallets to media.
    """
    _default_path = './wallet.json'

    def __init__(self,
                 path: str,
                 name: Optional[str] = None,
                 version: str = Wallet._wallet_version,
                 scrypt: Optional[ScryptParameters] = None,
                 accounts: List[Account] = None,
                 default_account: Optional[Account] = None,
                 extra: Optional[dict] = None):
        """

        Args:
            path: the location where the wallet will be stored.
            name: a user defined label for the wallet
            version: the wallet's version, must be equal to or greater than 3.0
            scrypt:  the parameters of the Scrypt algorithm used for encrypting and decrypting the private keys in the
            wallet.
            accounts: an array of Account objects to add to the wallet.
            extra: a user defined object for storing extra data. This field can be None.
        """

        filepath, extension = os.path.splitext(path)
        if len(extension) == 0:
            # if the path doesn't have a file extension, sets it as a .json file
            path += '.json'

        if name is None:
            # sets the wallet name the same as the file name
            dir_path, name = os.path.split(path)
            name, extension = os.path.splitext(name)

        self.path: str = path
        super().__init__(name=name,
                         version=version,
                         scrypt=scrypt,
                         accounts=accounts,
                         default_account=default_account,
                         extra=extra)

    def save(self) -> None:
        """
        Persists the wallet
        """
        with open(self.path, 'w') as json_file:
            json.dump(self.to_json(), json_file)

    @classmethod
    def default(cls, path: str = _default_path, name: Optional[str] = 'wallet.json') -> NEP6DiskWallet:
        """
        Create a new Wallet with the default settings.

        Args:
            path: the JSON's path.
            name: the Wallet name.
        """
        return cls(path=path,
                   name=name,
                   version=cls._wallet_version,
                   scrypt=ScryptParameters(),
                   accounts=[],
                   extra=None)



from __future__ import annotations
import asyncio
import aiodns  # type: ignore
import ipaddress
from neo3 import network_logger as logger, settings, blockchain, _Singleton
from neo3.network import node, payloads, convenience, message
from neo3.core import msgrouter
from typing import List, Optional
from contextlib import suppress
from datetime import datetime


def is_ip_address(hostname: str) -> bool:
    host = hostname.split(':')[0]
    try:
        ip = ipaddress.ip_address(host)
        return True
    except ValueError:
        return False


class NodeManager(_Singleton):
    """
    This class is a convenience class that helps establish and maintain a pool of active connections to NEO nodes.

    Attention:
        This class is singleton.
    """
    #: Time interval in seconds for asking nodes for their address list.
    ADDR_QUERY_INTERVAL = 15
    #: Time interval in seconds for calling the height monitoring check.
    MONITOR_HEIGHT_INTERVAL = 30
    #: Time interval in seconds for pinging remote nodes. Ping (and the Pong) response informs each other about chain
    #: heights
    PING_INTERVAL = 2

    #: Maximum time in seconds a node may take to update its best height before being removed for being stuck.
    MAX_HEIGHT_UPDATE_DURATION = 120

    #: Time interval in seconds to check the connection pool for open spots.
    POOL_CHECK_INTERVAL = 10
    #: Maximum number of time the minimum connected client setting may be violated by the open connect checker.
    MAX_NODE_POOL_ERROR = 2
    MAX_NODE_POOL_ERROR_COUNT = 0

    #: Maximum number of times adding a block or header may fail before the node is disconnected.
    MAX_NODE_ERROR_COUNT = 5

    #: Maximum number of times a timeout threshold may be exceeded before the node is disconnected.
    #: Requires calling :meth:`~neo3.network.convenience.nodemanager.NodeManager.increase_node_timeout_count` which is
    #: done automatically by the :class:`~neo3.network.convenience.syncmanager.SyncManager` if used.
    MAX_NODE_TIMEOUT_COUNT = 15

    # init() is used instead of __init__() due to the Singleton inheritance (read its class documentation)
    def init(self):
        #: A list of nodes that we're actively using to request data from
        self.nodes: List[node.NeoNode] = []
        #: A list of host:port addresses that have a task pending to to connect to, but are not fully processed
        self.queued_addresses = []

        self.tasks = []
        #: THe maximum number of clients to have connected at any time.
        self.max_clients = 5
        #: The minimum number clients to always have connected. If this is not met
        #: :attr:`~neo3.network.convenience.nodemanager.MAX_NODE_ERROR_COUNT` times then recovery logic will trigger.
        self.min_clients = 1
        self.shutting_down = False
        self.is_running = False

        msgrouter.on_node_connected += self._handle_node_connected
        msgrouter.on_node_disconnected += self._handle_node_disconnected

        self._test_client_provider = None

    def start(self) -> None:
        """
        Start the node manager services. This does 2 things

        1. Connect to the seed list addresses provided in the configuration
        2. Try to maintain a pool of connected nodes according to the `min/max clients` configuration settings and
           monitor that they don't get stuck.
        """
        def _start_services():
            self.is_running = True
            # 2. connect to addresses
            self._run_in_loop(self._fill_open_connection_spots, interval=self.POOL_CHECK_INTERVAL)

            # 3. keep trying to expand our address list with new addresses
            self._run_in_loop(self._query_addresses, interval=self.ADDR_QUERY_INTERVAL)

            # 4. monitor that connected nodes are not stuck
            self._run_in_loop(self._send_ping, interval=self.PING_INTERVAL)
            self._run_in_loop(self._monitor_node_height, interval=self.MONITOR_HEIGHT_INTERVAL)

        # 1. build an initial address list
        task = asyncio.create_task(self._process_seed_list_addresses())
        task.add_done_callback(lambda fut: _start_services())

    async def shutdown(self) -> None:
        """
        Gracefully shutdown the node manager.

        Disconnects all active nodes and stops service tasks.

        Note:
            This dependents on asyncio's Task canceling logic. It waits for all tasks to be cancelled and/or stopped
            before returning.
        """
        self.shutting_down = True

        # shutdown all running tasks for this class
        # to prevent automatic filling of open spots after disconnecting nodes
        logger.debug("stopping tasks...")
        for t in self.tasks:
            t.cancel()
        await asyncio.gather(*self.tasks, return_exceptions=True)

        # finally disconnect all existing connections
        # we need to create a new list to loop over, because `disconnect` removes items from self.nodes
        to_disconnect = list(map(lambda n: n, self.nodes))
        disconnect_tasks = []
        logger.debug("disconnecting nodes...")
        for n in to_disconnect:
            disconnect_tasks.append(asyncio.create_task(n.disconnect(payloads.DisconnectReason.SHUTTING_DOWN)))
        await asyncio.gather(*disconnect_tasks, return_exceptions=True)

    """
    Start utility functions
    """
    def get_node_addresses(self) -> List[payloads.NetworkAddress]:
        """
        Return the addresses of connected nodes.
        """
        return list(map(lambda n: n.address, self.nodes))

    def get_node_by_nodeid(self, nodeid: int) -> Optional[node.NeoNode]:
        """
        Return the node instance matching the provided nodeid.
        """
        for n in self.nodes:
            if n.nodeid == nodeid:
                return n
        else:
            return None

    def get_node_with_height(self, height: int) -> Optional[node.NeoNode]:
        """
        Return a node that has a chain height of at least `height`.

        Args:
            height: the minimal block height a node must have.

        Returns:
            None if there are no nodes in the current pool that match the criteria.
            Node instance sorted by weight to avoid hitting the same node every time.
        """
        if len(self.nodes) == 0:
            return None

        weights = list(map(lambda n: n.nodeweight, self.nodes))
        # highest weight is taken first
        weights.sort(reverse=True)

        for weight in weights:
            node = self.get_node_by_nodeid(weight.id)
            if node and height <= node.best_height:
                return node
        else:
            # we could not find a node with the height we're looking for
            return None

    def get_least_failed_node(self, ri: convenience.RequestInfo) -> Optional[node.NeoNode]:
        """
        Return the node with the least fail count for the target item as specified in the RequestInfo.

        Args:
            ri: the info object indicating which header or block we filter on.

        Returns:
            None if there are no connected nodes that can provide the target information as specified in the RequestInfo
            Node instance with the lowest failure count.
        """
        # Find the node with the least failures for the item in RequestInfo

        least_failed_times = 999
        least_failed_node = None
        tried_nodes = []

        for n in self.nodes:  # type: node.NeoNode
            if n.disconnecting:
                continue

            if n.best_height < ri.height:
                continue

            failed_times = ri.failed_nodes.get(n.nodeid, 0)
            if failed_times == 0:
                # return the node we haven't tried this request on before
                return n

            tried_nodes.append(n.nodeid)
            if failed_times < least_failed_times:
                least_failed_times = failed_times
                least_failed_node = n

        return least_failed_node

    def increase_node_error_count(self, nodeid: int) -> None:
        """
        Utility function to increase a node's `error_response_count` param by 1 and disconnect the node if it exceeds
        the threshold set by :attr:`~neo3.network.convenience.nodemanager.MAX_NODE_ERROR_COUNT`.

        Args:
            nodeid (:attr:`~neo3.network.node.NeoNode.nodeid`): the specific node to update.
        """
        node = self.get_node_by_nodeid(nodeid)
        if node:
            node.nodeweight.error_response_count += 1

            if node.nodeweight.error_response_count > self.MAX_NODE_ERROR_COUNT:
                logger.debug(f"Disconnecting node {node.nodeid} Reason: max error count threshold exceeded.")
                asyncio.create_task(node.disconnect(reason=payloads.DisconnectReason.POOR_PERFORMANCE))

    def increase_node_timeout_count(self, nodeid: int) -> None:
        """
        Utility function to increase a node's `timeout_count` param by 1 and disconnect the node if it exceeds the
        threshold set by :attr:`~neo3.network.convenience.nodemanager.MAX_NODE_ERROR_COUNT`.

        Args:
            nodeid (:attr:`~neo3.network.node.NeoNode.nodeid`): the specific node to update.
        """
        node = self.get_node_by_nodeid(nodeid)
        if node:
            node.nodeweight.timeout_count += 1

            if node.nodeweight.timeout_count > self.MAX_NODE_TIMEOUT_COUNT:
                logger.debug(f"Disconnecting node {node.nodeid_human} Reason: max timeout count threshold exceeded.")
                asyncio.create_task(node.disconnect(reason=payloads.DisconnectReason.POOR_PERFORMANCE))

    """
    End utility functions
    """

    """
    Start automatic service methods.

    This section contains the logic for establishing and maintaining node connections
    """
    def _handle_node_connected(self, node: node.NeoNode) -> None:
        """ Handle node connection event."""
        self.nodes.append(node)

        with suppress(ValueError):
            self.queued_addresses.remove(node.address)

    def _handle_node_disconnected(self, node: node.NeoNode, reason: payloads.DisconnectReason) -> None:
        """ Handle node disconnection event."""
        with suppress(ValueError):
            self.nodes.remove(node)

        with suppress(ValueError):
            self.queued_addresses.remove(node.address)

    async def _process_seed_list_addresses(self) -> None:
        """
        Parses addresses from the seed list and store valid addresses
        """
        resolver = aiodns.DNSResolver(loop=asyncio.get_event_loop(), timeout=2)

        async def _process(seed):
            host, port = seed.split(':')
            if not is_ip_address(host):
                try:
                    result = await resolver.query(host, 'A')
                    host = result[0].host
                except aiodns.error.DNSError as e:
                    logger.debug(f"Skipping {host}, address could not be resolved: {e}.")
            node.NeoNode.addresses.append(payloads.NetworkAddress(address=f"{host}:{port}"))

        tasks = []
        for seed in settings.network.seedlist:
            tasks.append(_process(seed))
        await asyncio.gather(*tasks)

    def _connect_done_cb(self, future) -> None:
        node_instance, failure = future.result()
        # failures here are hard failures from asyncio's loop.create_connection()
        if failure:
            logger.debug(f"Failed to connect to {failure[0]} reason: {failure[1]}.")
            tmp_addr = payloads.NetworkAddress(address=failure[0])

            with suppress(ValueError):
                idx = node.NeoNode.addresses.index(tmp_addr)
                addr = node.NeoNode.addresses[idx]
                addr.set_state_dead()
                self.queued_addresses.remove(tmp_addr)

            msgrouter.on_client_connect_done(None, failure)
        else:
            msgrouter.on_client_connect_done(node_instance, None)
            node_instance.start_message_handler()

        self.tasks.remove(future)

    async def _fill_open_connection_spots(self) -> None:
        open_spots = self.max_clients - (len(self.nodes) + len(self.queued_addresses))

        if open_spots > 0:
            logger.debug(f"Found {open_spots} open pool spots, trying to add nodes...")

            # we sort the addresses such that nodes we recently disconnected from are last in the list
            # this matters in case we had to recycle addresses, meaning addresses with state POOR
            # are now labelled NEW again.
            node.NeoNode.addresses.sort(key=lambda addr: addr.last_connected)

            for _ in range(open_spots):
                # now we ask for the first address with the state NEW
                addr = node.NeoNode.get_address_new()
                if addr:
                    # an address can be queued and its state not yet changed to CONNECTED, so we must make sure we're
                    # not trying to connect to an address that is in an ongoing connection state
                    if addr not in self.queued_addresses:
                        logger.debug(f"Adding {addr} to connection queue.")
                        self.queued_addresses.append(addr)
                        if self._test_client_provider:
                            socket_mock = next(self._test_client_provider())
                            task = asyncio.create_task(node.NeoNode.connect_to(socket=socket_mock))
                        else:
                            task = asyncio.create_task(node.NeoNode.connect_to(addr.ip, addr.port))
                        self.tasks.append(task)
                        task.add_done_callback(self._connect_done_cb)
                else:
                    # oh no, we've exhausted our NEW addresses list
                    if len(self.nodes) >= self.min_clients:
                        logger.debug(f"No addresses available to fill spots. However, minimum clients still satisfied.")
                        break
                    else:
                        if self.MAX_NODE_POOL_ERROR_COUNT != self.MAX_NODE_POOL_ERROR:
                            # give our `_query_addresses` loop a chance to collect new addresses from connected nodes
                            self.MAX_NODE_POOL_ERROR_COUNT += 1
                            logger.debug(f"Increasing pool spot error count to {self.MAX_NODE_POOL_ERROR_COUNT}.")
                            break
                        else:
                            # we have no other option then to retry any address we know
                            logger.debug("Recycling old addresses.")
                            for addr in node.NeoNode.addresses:
                                if addr.is_state_poor:
                                    addr.set_state_new()
                            self.MAX_NODE_POOL_ERROR_COUNT = 0
                            break

    async def _query_addresses(self) -> None:
        """
        Ask for the address list of connected nodes on an interval.
        """
        logger.debug(f"Connected node count {len(self.nodes)}.")
        for node in self.nodes:
            logger.debug(f"Asking node {node.nodeid_human} for its address list")
            task = asyncio.create_task(node.request_address_list())
            self.tasks.append(task)
            task.add_done_callback(lambda fut: self.tasks.remove(fut))

    async def _monitor_node_height(self) -> None:
        now = datetime.utcnow().timestamp()
        for node in self.nodes:
            if now - node.best_height_last_update > self.MAX_HEIGHT_UPDATE_DURATION:
                logger.debug(f"Disconnecting node {node.nodeid} Reason: max height update threshold exceeded.")
                asyncio.create_task(node.disconnect(reason=payloads.DisconnectReason.POOR_PERFORMANCE))
            else:
                logger.debug(f"Asking node {node.nodeid_human} to send us a height update (PING)")
                # Request latest height from node
                if settings.database:
                    height = max(0, blockchain.Blockchain().height)
                else:
                    height = 0
                m = message.Message(msg_type=message.MessageType.PING, payload=payloads.PingPayload(height=height))
                task = asyncio.create_task(node.send_message(m))
                self.tasks.append(task)
                task.add_done_callback(lambda fut: self.tasks.remove(fut))

    async def _send_ping(self):
        """
        Inform connected nodes of our height while also learning if their height has changed

        NEO C# Preview3 no longer seems to instantly inform its connected nodes about new available blocks. Therefore
        syncing can be 2-3 blocks behind. By calling this in an interval we improve
        1) new block receiving speed when our local chain is "in sync".
        2) our node height monitoring service
        """
        for remote_node in self.nodes:  # type: node.NeoNode
            asyncio.create_task(remote_node.send_ping())

    def _run_in_loop(self, coro_func, interval) -> None:
        """
        Helper function to run a coroutine every `interval` seconds as long as the node manager is running.

        Stores a task in `tasks`

        Args:
            coro:
            interval:
        """
        async def _t(coro_func, interval):
            while not self.shutting_down:
                await coro_func()
                await asyncio.sleep(interval)

        self.tasks.append(asyncio.create_task(_t(coro_func, interval)))

    """
    End automatic service methods.
    """

    def _reset_for_test(self) -> None:
        self.nodes = []
        self.queued_addresses = []
        self.tasks = []
        self._test_client_provider = None
        self.is_running = False



from __future__ import annotations
import base64
from typing import TYPE_CHECKING
from enum import IntEnum
from neo3.network import payloads
from neo3.core import Size as s, utils, serialization, IJson
from neo3 import vm, contracts

if TYPE_CHECKING:
    from neo3 import storage


class OracleReponseCode(IntEnum):
    SUCCESS = 0x00
    PROTOCOL_NOT_SUPPORTED = 0x10
    CONSENSUS_UNREACHABLE = 0x12
    NOT_FOUND = 0x14
    TIMEOUT = 0x16
    FORBIDDEN = 0x18
    RESPONSE_TOO_LARGE = 0x1a
    INSUFFICIENT_FUNDS = 0x1c
    CONTENT_TYPE_NOT_SUPPORTED = 0x1f
    ERROR = 0xFF


class OracleResponse(payloads.TransactionAttribute, IJson):
    _MAX_RESULT_SIZE = 0xFFFF
    _FIXED_ORACLE_SCRIPT = None

    def __init__(self, id: int, code: OracleReponseCode, result: bytes):
        super(OracleResponse, self).__init__()
        self.type_ = payloads.TransactionAttributeType.ORACLE_RESPONSE
        #: Only one OracleResponse attribute can be attached per transaction
        self.allow_multiple = False
        #: The OracleRequest id to which this is a response
        self.id = id
        #: The evaluation result code
        self.code = code
        #: The actual result
        self.result = result
        if self._FIXED_ORACLE_SCRIPT is None:
            sb = vm.ScriptBuilder()
            sb.emit_dynamic_call(contracts.OracleContract().hash, "finish")  # type: ignore
            self._FIXED_ORACLE_SCRIPT = sb.to_array()

    def __len__(self):
        return super(OracleResponse, self).__len__() + s.uint64 + s.uint8 + utils.get_var_size(self.result)

    def _deserialize_without_type(self, reader: serialization.BinaryReader) -> None:
        self.id = reader.read_uint64()
        self.code = OracleReponseCode(reader.read_uint8())
        self.result = reader.read_var_bytes(self._MAX_RESULT_SIZE)
        if self.code != OracleReponseCode.SUCCESS and len(self.result) > 0:
            raise ValueError(f"Deserialization error - oracle response: {self.code}")

    def _serialize_without_type(self, writer: serialization.BinaryWriter) -> None:
        writer.write_uint64(self.id)
        writer.write_uint8(self.code)
        writer.write_var_bytes(self.result)

    def to_json(self) -> dict:
        """ Convert object into json """
        json = super(OracleResponse, self).to_json()
        json.update({"id": id,
                     "code": self.code,
                     "result": base64.b64encode(self.result)}
                    )
        return json

    @classmethod
    def from_json(cls, json: dict):
        """ Create object from JSON """
        return cls(json['id'], json['code'], base64.b64decode(json['result']))

    def verify(self, snapshot: storage.Snapshot, tx: payloads.Transaction) -> bool:
        """
        Verifies the attribute with the transaction

        Returns:
            True if verification passes. False otherwise.
        """
        if any(map(lambda signer: signer.scope != payloads.WitnessScope.NONE, tx.signers)):
            return False

        if tx.script != self._FIXED_ORACLE_SCRIPT:
            return False

        oracle = contracts.native.OracleContract()
        request = oracle.get_request(snapshot, self.id)
        if request is None:
            return False
        if tx.network_fee + tx.system_fee != request.gas_for_response:
            return False
        oracle_account = contracts.Contract.get_consensus_address(
            contracts.DesignationContract().get_designated_by_role(snapshot,
                                                                   contracts.DesignateRole.ORACLE,
                                                                   snapshot.best_block_height + 1)
        )
        return any(map(lambda signer: signer.account == oracle_account, tx.signers))

    @classmethod
    def _serializable_init(cls):
        return cls(0, OracleReponseCode.ERROR, b'')



from __future__ import annotations
import mmh3  # type: ignore
from neo3 import vm, contracts, settings
from neo3.core import cryptography, IInteroperable, types, msgrouter, to_script_hash
from neo3.contracts.interop import register


@register("System.Runtime.Platform", 1 << 3, contracts.CallFlags.NONE)
def get_platform(engine: contracts.ApplicationEngine) -> str:
    return "NEO"


@register("System.Runtime.GetNetwork", 1 << 3, contracts.CallFlags.NONE)
def get_network(engine: contracts.ApplicationEngine) -> str:
    return settings.network.magic


@register("System.Runtime.GetRandom", 1 << 4, contracts.CallFlags.NONE)
def get_random(engine: contracts.ApplicationEngine) -> int:
    engine.nonce_data = mmh3.hash_bytes(engine.nonce_data, settings.network.magic)
    return int.from_bytes(engine.nonce_data, "little", signed=False)


@register("System.Runtime.GetTrigger", 1 << 3, contracts.CallFlags.NONE)
def get_trigger(engine: contracts.ApplicationEngine) -> vm.BigInteger:
    return vm.BigInteger(engine.trigger.value)


@register("System.Runtime.GetTime", 1 << 3, contracts.CallFlags.NONE)
def get_time(engine: contracts.ApplicationEngine) -> int:
    return engine.snapshot.persisting_block.timestamp


@register("System.Runtime.GetScriptContainer", 1 << 3, contracts.CallFlags.NONE)
def get_scriptcontainer(engine: contracts.ApplicationEngine) -> vm.StackItem:
    if not isinstance(engine.script_container, IInteroperable):
        raise ValueError("script container is not a valid IInteroperable type")
    return engine.script_container.to_stack_item(engine.reference_counter)


@register("System.Runtime.GetExecutingScriptHash", 1 << 4, contracts.CallFlags.NONE)
def get_executingscripthash(engine: contracts.ApplicationEngine) -> types.UInt160:
    return engine.current_scripthash


@register("System.Runtime.GetCallingScriptHash", 1 << 4, contracts.CallFlags.NONE)
def get_callingscripthash(engine: contracts.ApplicationEngine) -> types.UInt160:
    return engine.calling_scripthash


@register("System.Runtime.GetEntryScriptHash", 1 << 4, contracts.CallFlags.NONE)
def get_entryscripthash(engine: contracts.ApplicationEngine) -> types.UInt160:
    return engine.entry_scripthash


@register("System.Runtime.CheckWitness", 1 << 10, contracts.CallFlags.NONE)
def do_checkwitness(engine: contracts.ApplicationEngine, data: bytes) -> bool:
    if len(data) == 20:
        hash_ = types.UInt160(data)
    elif len(data) == 33:
        redeemscript = contracts.Contract.create_signature_redeemscript(
            cryptography.ECPoint.deserialize_from_bytes(data)
        )
        hash_ = to_script_hash(redeemscript)
    else:
        raise ValueError("Supplied CheckWitness data is not a valid hash")

    return engine.checkwitness(hash_)


@register("System.Runtime.GetInvocationCounter", 1 << 4, contracts.CallFlags.NONE)
def get_invocationcounter(engine: contracts.ApplicationEngine) -> int:
    return engine.get_invocation_counter()


@register("System.Runtime.Log", 1 << 15, contracts.CallFlags.ALLOW_NOTIFY)
def do_log(engine: contracts.ApplicationEngine, message: bytes) -> None:
    if len(message) > engine.MAX_NOTIFICATION_SIZE:
        raise ValueError(f"Log message length ({len(message)}) exceeds maximum allowed ({engine.MAX_NOTIFICATION_SIZE})")  # noqa
    msgrouter.interop_log(engine.script_container, message.decode('utf-8'))


@register("System.Runtime.Notify", 1 << 15, contracts.CallFlags.ALLOW_NOTIFY)
def do_notify(engine: contracts.ApplicationEngine, event_name: bytes, state: vm.ArrayStackItem) -> None:
    """

    Args:
        engine:
        event_name:
        state: values belonging to the notification event.
        e.g. a NEP-17 transfer event might have as state: from script_hash, to script_hash and an ammount
    """
    if len(event_name) > engine.MAX_EVENT_SIZE:
        raise ValueError(
            f"Notify event name length ({len(event_name)}) exceeds maximum allowed ({engine.MAX_EVENT_SIZE})")
    # will validate size + cyclic references
    contracts.BinarySerializer.serialize(state, engine.MAX_NOTIFICATION_SIZE)
    engine.notifications.append((engine.script_container, engine.current_scripthash, event_name, state))
    msgrouter.interop_notify(engine.current_scripthash, event_name.decode('utf-8'), state)


@register("System.Runtime.GetNotifications", 1 << 8, contracts.CallFlags.NONE)
def get_notifications(engine: contracts.ApplicationEngine, for_hash: types.UInt160) -> vm.ArrayStackItem:
    array = vm.ArrayStackItem(engine.reference_counter)
    for notification in engine.notifications:
        if notification[1] == for_hash:
            notification_stackitem = vm.ArrayStackItem(engine.reference_counter)
            notification_stackitem.append([
                vm.ByteStringStackItem(notification[1].to_array()),  # script_hash
                vm.ByteStringStackItem(notification[2]),  # message
                notification[3].deep_copy()  # state
            ])
            array.append(notification_stackitem)
    if len(array) > engine.MAX_STACK_SIZE:
        raise ValueError("Notification count exceeds limits")
    return array


@register("System.Runtime.GasLeft", 1 << 4, contracts.CallFlags.NONE)
def get_gasleft(engine: contracts.ApplicationEngine) -> int:
    if engine.is_test_mode:
        return -1
    else:
        return engine.gas_amount - engine.gas_consumed


@register("System.Runtime.BurnGas", 1 << 4, contracts.CallFlags.NONE)
def do_burn_gas(engine: contracts.ApplicationEngine, gas: int) -> None:
    if gas <= 0:
        raise ValueError("Burn gas cannot be called with negative value")
    engine.add_gas(gas)



from __future__ import annotations
from typing import List
from datetime import datetime, timezone
from neo3 import contracts, storage, settings, vm, _Singleton
from neo3.core import cryptography, types, to_script_hash, msgrouter, syscall_name_to_int
from neo3.network import payloads, convenience


class Blockchain(_Singleton):
    genesis_block = None

    def init(self, backend: storage.IDBImplementation = None, store_genesis_block=True):  # type: ignore
        self.backend = backend if backend else settings.database
        self._current_snapshot = None
        self.genesis_block = self._create_genesis_block()
        self.genesis_block.rebuild_merkle_root()

        sb = vm.ScriptBuilder().emit_syscall(syscall_name_to_int("System.Contract.NativeOnPersist"))
        self.native_onpersist_script = sb.to_array()
        sb = vm.ScriptBuilder().emit_syscall(syscall_name_to_int("System.Contract.NativePostPersist"))
        self.native_postpersist_script = sb.to_array()

        if self.currentSnapshot.best_block_height < 0 and store_genesis_block:
            self.persist(self.genesis_block)

    @property
    def currentSnapshot(self) -> storage.Snapshot:
        if self._current_snapshot is None:
            self._current_snapshot = self.backend.get_snapshotview()
        return self._current_snapshot  # type: ignore
        # mypy thinks it returns None

    @property
    def height(self):
        return self.currentSnapshot.best_block_height

    @staticmethod
    def _create_genesis_block() -> payloads.Block:
        h = payloads.Header(
            version=0,
            prev_hash=types.UInt256.zero(),
            timestamp=int(datetime(2016, 7, 15, 15, 8, 21, 0, timezone.utc).timestamp() * 1000),
            nonce=2083236893,
            index=0,
            primary_index=0,
            next_consensus=contracts.Contract.get_consensus_address(settings.standby_validators),
            witness=payloads.Witness(
                invocation_script=b'',
                verification_script=b'\x11'  # (OpCode.PUSH1)
            ),
        )
        return payloads.Block(header=h, transactions=[])

    def persist(self, block: payloads.Block):
        with self.backend.get_snapshotview() as snapshot:
            snapshot.persisting_block = block

            engine = contracts.ApplicationEngine(contracts.TriggerType.ON_PERSIST,
                                                 None, snapshot, 0, True)  # type: ignore
            engine.load_script(vm.Script(self.native_onpersist_script))
            if engine.execute() != vm.VMState.HALT:
                raise ValueError(f"Failed onPersist in native contracts: {engine.exception_message}")

            cloned_snapshot = snapshot.clone()
            for tx in block.transactions:
                tx.block_height = block.index
                cloned_snapshot.transactions.put(tx)
                cloned_snapshot.transactions.commit()

                engine = contracts.ApplicationEngine(contracts.TriggerType.APPLICATION,
                                                     tx,
                                                     cloned_snapshot,
                                                     tx.system_fee)
                engine.load_script(vm.Script(tx.script))
                state = engine.execute()
                if state == vm.VMState.HALT:
                    cloned_snapshot.commit()
                else:
                    cloned_snapshot = snapshot.clone()

            engine = contracts.ApplicationEngine(contracts.TriggerType.POST_PERSIST,
                                                 None, snapshot, 0, True)  # type: ignore
            engine.load_script(vm.Script(self.native_postpersist_script))
            if engine.execute() != vm.VMState.HALT:
                raise ValueError(f"Failed postPersist in native contracts: {engine.exception_message}")
            """
            LedgerContract updates the current block in the post_persist event
            this means transactions in the persisting block that call LedgerContract.current_hash/current_index()
            will get refer the (previous) block hash/index, not the block they're included in.

            Therefore we wait with persisting the block until here
            """
            snapshot.blocks.put(block)
            snapshot.best_block_height = block.index

            snapshot.commit()
            self._current_snapshot = snapshot
        msgrouter.on_block_persisted(block)



from __future__ import annotations
from neo3 import contracts


def register(method: str,
             flags: contracts.CallFlags,
             *,
             cpu_price: int = 0,
             storage_price: int = 0):
    """
    Register a publicly callable method on a native contract

    Args:
        method: name of call.
        cpu_price: the computational price of calling the handler.
        storage_price: the storage price of calling the handler.
        flags: ExecutionContext rights needed.
    """
    def inner_func(func):
        func.native_call = True
        func.name = method
        func.cpu_price = cpu_price
        func.storage_price = storage_price
        func.flags = flags
        return func
    return inner_func



import unittest
from neo3 import contracts
from neo3.core import types


class ContractStateTestCase(unittest.TestCase):
    def test_contractstate_clone(self):
        manifest = contracts.ContractManifest()
        nef = contracts.NEF()
        state = contracts.ContractState(1, nef, manifest, 0, types.UInt160.zero())
        clone = state.clone()
        self.assertNotEqual(id(state), id(clone))
        self.assertNotEqual(id(state.manifest), id(clone.manifest))

    def test_equals(self):
        manifest = contracts.ContractManifest()
        nef = contracts.NEF()
        state = contracts.ContractState(1, nef, manifest, 0, types.UInt160.zero())
        clone = state.clone()
        self.assertEqual(state, clone)

        nef2 = contracts.NEF()
        state2 = contracts.ContractState(2, nef2, manifest, 0, types.UInt160(b'\x01' * 20))
        self.assertNotEqual(state, state2)
        self.assertNotEqual(state, None)
        self.assertNotEqual(state, object())



from __future__ import annotations
from datetime import datetime
from random import randint
from neo3.core import Size as s
from neo3.core import serialization


class PingPayload(serialization.ISerializable):
    def __init__(self, height: int = 0) -> None:
        #: The current local chain height
        self.current_height = height
        #: The local time in UTC as a timestamp
        self.timestamp = int(datetime.utcnow().timestamp())
        #: Random number
        self.nonce = randint(100, 10000)

    def __len__(self) -> int:
        """ Get the total size in bytes of the object. """
        return s.uint32 + s.uint32 + s.uint32

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.current_height)
        writer.write_uint32(self.timestamp)
        writer.write_uint32(self.nonce)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.current_height = reader.read_uint32()
        self.timestamp = reader.read_uint32()
        self.nonce = reader.read_uint32()



import json
import unittest
from neo3 import contracts, vm


class JSONParsingTestCase(unittest.TestCase):
    def _parse_and_dump(self, input):
        return contracts.NEOJson.dumps(contracts.NEOJson.loads(input))

    def test_wrong_json(self):
        data = [
            "[    ]XXXXXXX",
            "{   }XXXXXXX",
            "[,,,,]",
            "false,X",
            "false@@@",
            "{""length"":99999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999}"  # noqa
        ]

        for entry in data:
            with self.assertRaises(json.decoder.JSONDecodeError):
                self._parse_and_dump(entry)

    def test_array(self):
        self.assertEqual("[]", self._parse_and_dump("[    ]"))

        expected = "[1,\"a==\",-1.3,null]"
        self.assertEqual(expected, self._parse_and_dump("[1,\"a==\",    -1.3 ,null] "))

    def test_bool(self):
        expected = "[true,false]"
        self.assertEqual(expected, self._parse_and_dump("[  true ,false ]"))

        with self.assertRaises(json.decoder.JSONDecodeError):
            self._parse_and_dump("[True,FALSE] ")

    def test_numbers(self):
        expected = "[1,-2,3.5]"
        self.assertEqual(expected, self._parse_and_dump("[  1, -2 , 3.5 ]"))

        expected = "[20050000,20050000,-1.1234e-100]"
        self.assertEqual(expected, self._parse_and_dump("[200.500000E+005,200.500000e+5,-1.1234e-100]"))

        data = [
            "[-]",
            "[1.]",
            "[.123]",
            "[--1.123]",
            "[+1.123]",
            "[1.12.3]",
            "[e--1]",
            "[e++1]",
            "[E- 1]",
            "[3e--1]",
            "[2e++1]",
            "[1E- 1]",
        ]

        for entry in data:
            with self.assertRaises(json.decoder.JSONDecodeError):
                self._parse_and_dump(entry)

    def test_string(self):
        expected = r'["","\b\f\t\n\r/\\"]'
        self.assertEqual(expected, self._parse_and_dump(r' ["" ,  "\b\f\t\n\r\/\\" ]'))

        expected = r'["\uD834\uDD1E"]'
        self.assertEqual(expected.lower(), self._parse_and_dump(expected))

        expected = r'["\\x00"]'
        self.assertEqual(expected, self._parse_and_dump(expected))

        data = [
            r'["]',
            r'[""\uaaa""]',
            r'["\uaa"]',
            r'["\ua"]',
            r'["\u"]',
        ]

        for entry in data:
            with self.assertRaises(json.decoder.JSONDecodeError):
                self._parse_and_dump(entry)

    def test_object(self):
        expected = "{\"test\":true}"
        self.assertEqual(expected, self._parse_and_dump(" {\"test\":   true}"))

        expected = r'{"\uaaaa":true}'
        self.assertEqual(expected, self._parse_and_dump(r' {"\uAAAA":   true}'))

        data = [
            "{\"a\":}",
            "{NULL}",
            "[\"a\":]"
        ]

        for entry in data:
            with self.assertRaises(json.JSONDecodeError):
                self._parse_and_dump(entry)

    def test_max_depth(self):
        data = r'{"1":{"2":2}}'
        with self.assertRaises(json.JSONDecodeError) as context:
            contracts.NEOJson.loads(data, max_depth=1)
        self.assertEqual("Maximum depth exceeded", str(context.exception))

        data = "[[1]]"
        with self.assertRaises(json.JSONDecodeError) as context:
            contracts.NEOJson.loads(data, max_depth=1)
        self.assertEqual("Maximum depth exceeded", str(context.exception))

        data = r'{"1":[[2]]}'
        with self.assertRaises(json.JSONDecodeError) as context:
            contracts.NEOJson.loads(data, max_depth=2)
        self.assertEqual("Maximum depth exceeded", str(context.exception))

    def test_duplicate_keys(self):
        # By default Python accepts this and just takes the last key found
        # NEO doesn't and throws an error
        data = r'{"a":123, "a":456}'
        with self.assertRaises(json.JSONDecodeError) as context:
            contracts.NEOJson.loads(data)
        self.assertEqual("Duplicate keys in objects are not allowed", str(context.exception))


class JSONSerializerTestCase(unittest.TestCase):
    def test_deserialization_basics(self):
        # test empty object
        item = contracts.JSONSerializer.deserialize(contracts.NEOJson.loads("{}"), vm.ReferenceCounter())
        self.assertIsInstance(item, vm.MapStackItem)
        self.assertEqual(0, len(item))

        # test empty array
        item = contracts.JSONSerializer.deserialize(contracts.NEOJson.loads("[]"), vm.ReferenceCounter())
        self.assertIsInstance(item, vm.ArrayStackItem)
        self.assertEqual(0, len(item))

    def test_deserialization_map(self):
        ref_ctr = vm.ReferenceCounter()
        item = contracts.JSONSerializer.deserialize(contracts.NEOJson.loads(r'{"test1":123, "test2": 321}'), ref_ctr)
        self.assertIsInstance(item, vm.MapStackItem)
        self.assertEqual(2, len(item))

        key1 = vm.ByteStringStackItem("test1")
        self.assertEqual(vm.BigInteger(123), item[key1].to_biginteger())
        key2 = vm.ByteStringStackItem("test2")
        self.assertEqual(vm.BigInteger(321), item[key2].to_biginteger())

    def test_deserialization_array_of_items(self):
        ref_ctr = vm.ReferenceCounter()
        array = contracts.JSONSerializer.deserialize(
            contracts.NEOJson.loads(
                r'[[true,"test1", 123, null],[false,"test2",321]]'
            ),
            ref_ctr
        )
        self.assertIsInstance(array, vm.ArrayStackItem)
        self.assertEqual(2, len(array))

        sub_array1 = array[0]
        self.assertIsInstance(sub_array1, vm.ArrayStackItem)
        self.assertEqual(4, len(sub_array1))

        self.assertIsInstance(sub_array1[0], vm.BooleanStackItem)
        self.assertTrue(sub_array1[0])

        self.assertIsInstance(sub_array1[1], vm.ByteStringStackItem)
        self.assertEqual(vm.ByteStringStackItem("test1"), sub_array1[1])

        self.assertIsInstance(sub_array1[2], vm.IntegerStackItem)
        self.assertEqual(vm.BigInteger(123), sub_array1[2].to_biginteger())

        self.assertIsInstance(sub_array1[3], vm.NullStackItem)

        sub_array2 = array[1]
        self.assertIsInstance(sub_array2, vm.ArrayStackItem)
        self.assertEqual(3, len(sub_array2))

        self.assertIsInstance(sub_array2[0], vm.BooleanStackItem)
        self.assertFalse(sub_array2[0])

        self.assertIsInstance(sub_array2[1], vm.ByteStringStackItem)
        self.assertEqual(vm.ByteStringStackItem("test2"), sub_array2[1])

        self.assertIsInstance(sub_array2[2], vm.IntegerStackItem)
        self.assertEqual(vm.BigInteger(321), sub_array2[2].to_biginteger())

    def test_deserialization_invalid_arguments(self):
        # invalid json_data, float is not part of JObject
        with self.assertRaises(ValueError):
            contracts.JSONSerializer.deserialize(float(1.2))

        with self.assertRaises(ValueError) as context:
            contracts.JSONSerializer.deserialize([1])
        self.assertEqual("Can't deserialize JSON array without reference counter", str(context.exception))

        with self.assertRaises(ValueError) as context:
            contracts.JSONSerializer.deserialize({})
        self.assertEqual("Can't deserialize JSON object without reference counter", str(context.exception))

    def test_serialization_basics(self):
        ref = vm.ReferenceCounter()
        m = vm.MapStackItem(ref)
        s = contracts.JSONSerializer.serialize(m, 999)
        self.assertEqual("{}", s)

        a = vm.ArrayStackItem(ref)
        s = contracts.JSONSerializer.serialize(a, 999)
        self.assertEqual(r'[]', s)

        i1 = vm.IntegerStackItem(1)
        i2 = vm.IntegerStackItem(9007199254740992)
        a.append([i1, i2])
        s = contracts.JSONSerializer.serialize(a, 999)
        self.assertEqual(r'[1,"9007199254740992"]', s)

    def test_serialization_map(self):
        ref = vm.ReferenceCounter()
        key1 = vm.ByteStringStackItem("test1")
        key2 = vm.ByteStringStackItem("test2")
        key3 = vm.ByteStringStackItem("test3")
        v1 = vm.IntegerStackItem(1)
        v2 = vm.IntegerStackItem(2)
        v3 = vm.IntegerStackItem(3)
        m = vm.MapStackItem(ref)
        m[key1] = v1
        m[key3] = v3
        m[key2] = v2
        s = contracts.JSONSerializer.serialize(m, 999)
        expected = r'{"test1":1,"test3":3,"test2":2}'
        self.assertEqual(expected, s)

    def test_serialization_array(self):
        b = vm.BooleanStackItem(True)
        bs = vm.ByteStringStackItem("test")
        i = vm.IntegerStackItem(123)
        n = vm.NullStackItem()
        ref_ctr = vm.ReferenceCounter()
        a = vm.ArrayStackItem(ref_ctr)
        a.append([b, bs, i, n])
        expected = r'[true,"test",123,null]'
        self.assertEqual(expected, contracts.JSONSerializer.serialize(a, 999))

    def test_serialization_array_nested(self):
        bool_t = vm.BooleanStackItem(True)
        bool_f = vm.BooleanStackItem(False)
        bs1 = vm.ByteStringStackItem("test1")
        bs2 = vm.ByteStringStackItem("test2")

        i1 = vm.IntegerStackItem(123)
        i2 = vm.IntegerStackItem(321)
        ref_ctr = vm.ReferenceCounter()

        a1 = vm.ArrayStackItem(ref_ctr)
        a1.append([bool_t, bs1, i1])

        a2 = vm.ArrayStackItem(ref_ctr)
        a2.append([bool_f, bs2, i2])

        parent = vm.ArrayStackItem(ref_ctr)
        parent.append([a1, a2])
        expected = r'[[true,"test1",123],[false,"test2",321]]'
        self.assertEqual(expected, contracts.JSONSerializer.serialize(parent, 999))

    def test_serialization_invalid_type(self):
        s = vm.Script(b'')
        p = vm.PointerStackItem(s, 0)
        with self.assertRaises(TypeError) as context:
            contracts.JSONSerializer.serialize(p, 999)
        self.assertEqual("Object of type PointerStackItem is not JSON serializable", str(context.exception))

    def test_serialization_max_size(self):
        data = [True]
        expected = "[true]"
        s = contracts.JSONSerializer.serialize(data, max_size=len(expected))
        self.assertEqual(expected, s)

        with self.assertRaises(ValueError):
            contracts.JSONSerializer.serialize(data, max_size=len(expected) - 1)

    def test_serialization_map_with_integer_key(self):
        i = vm.IntegerStackItem(123)
        v = vm.IntegerStackItem(321)
        ref_ctr = vm.ReferenceCounter()
        m = vm.MapStackItem(ref_ctr)
        m[i] = v
        expected = r'{"123":321}'
        s = contracts.JSONSerializer.serialize(m, 999)
        self.assertEqual(expected, s)





from __future__ import annotations
import enum
from typing import List, Optional, Type, Union, cast
from enum import IntEnum
from neo3.core import IJson, IInteroperable, serialization, cryptography
from neo3 import contracts, vm


class ContractParameterType(IntEnum):
    ANY = 0x00,
    BOOLEAN = 0x10,
    INTEGER = 0x11,
    BYTEARRAY = 0x12,
    STRING = 0x13,
    HASH160 = 0x14,
    HASH256 = 0x15,
    PUBLICKEY = 0x16,
    SIGNATURE = 0x17,
    ARRAY = 0x20,
    MAP = 0x22,
    INTEROPINTERFACE = 0x30,
    VOID = 0xff

    def PascalCase(self) -> str:
        if self == ContractParameterType.BYTEARRAY:
            return "ByteArray"
        elif self == ContractParameterType.INTEROPINTERFACE:
            return "InteropInterface"
        elif self == ContractParameterType.PUBLICKEY:
            return "PublicKey"
        else:
            return self.name.title()

    @classmethod
    def from_type(cls, class_type: Optional[Type[object]]) -> ContractParameterType:
        if class_type is None or class_type == type(None):
            return ContractParameterType.VOID
        elif class_type in [bool, vm.BooleanStackItem]:
            return ContractParameterType.BOOLEAN
        elif class_type in [int, vm.BigInteger]:
            return ContractParameterType.INTEGER
        elif class_type in [bytes, bytearray, vm.BufferStackItem, vm.ByteStringStackItem]:
            return ContractParameterType.BYTEARRAY
        elif class_type == cryptography.ECPoint:
            return ContractParameterType.PUBLICKEY
        elif hasattr(class_type, '__origin__'):
            if class_type.__origin__ == list:  # type: ignore
                return ContractParameterType.ARRAY
            if class_type.__origin__ == Union:  # type: ignore
                # handle typing.Optional[type], Optional is an alias for Union[x, None]
                # only support specifying 1 type
                if len(class_type.__args__) != 2:  # type: ignore
                    raise ValueError(f"Don't know how to convert {class_type}")
                for i in class_type.__args__:  # type: ignore
                    if i is None:
                        continue
                    return cls.from_type(i)
            raise ValueError
        elif issubclass(class_type, serialization.ISerializable):
            return ContractParameterType.BYTEARRAY
        elif class_type == str:
            return ContractParameterType.STRING
        elif class_type == vm.MapStackItem:
            return ContractParameterType.MAP
        elif class_type in [vm.ArrayStackItem, vm.StructStackItem, list]:
            return ContractParameterType.ARRAY
        elif issubclass(class_type, IInteroperable):
            return ContractParameterType.ARRAY
        elif class_type == vm.StackItem:
            return ContractParameterType.ANY
        elif issubclass(class_type, enum.Enum):
            return ContractParameterType.INTEGER
        else:
            return ContractParameterType.INTEROPINTERFACE


class ContractParameterDefinition(IJson):
    """
    A parameter description for a contract Method or Event.
    """
    def __init__(self, name: str, type: contracts.ContractParameterType):
        """
        Args:
            name: the human readable identifier.
            type: the type of parameter.
        """
        self.name = name
        self.type = type

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.name == other.name and self.type == other.type

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            "name": self.name,
            "type": self.type.PascalCase()
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractParameterDefinition:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the manifest name property has an incorrect format.
            ValueError: if the type is VOID.
        """
        c = cls(
            name=contracts.validate_type(json['name'], str),
            type=contracts.ContractParameterType[contracts.validate_type(json['type'], str).upper()]
        )
        if c.name is None or len(c.name) == 0:
            raise ValueError("Format error - invalid 'name'")
        if c.type == contracts.ContractParameterType.VOID:
            raise ValueError("Format error - parameter type VOID is not allowed")
        return c

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        return vm.StructStackItem(reference_counter,
                                  [vm.ByteStringStackItem(self.name), vm.IntegerStackItem(self.type.value)]
                                  )


class ContractEventDescriptor(IJson):
    """
    A description for an event that a contract can broadcast.
    """
    def __init__(self, name: str, parameters: List[ContractParameterDefinition]):
        """

        Args:
            name: the human readable identifier of the event.
            parameters: the list of parameters the event takes.
        """
        self.name = name
        self.parameters = parameters

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.name == other.name and self.parameters == other.parameters

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            "name": self.name,
            "parameters": list(map(lambda p: p.to_json(), self.parameters))
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractEventDescriptor:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary key.
            ValueError: if the 'name' property has an incorrect format
        """
        c = cls(
            name=contracts.validate_type(json['name'], str),
            parameters=list(map(lambda p: ContractParameterDefinition.from_json(p), json['parameters']))
        )
        if c.name is None or len(c.name) == 0:
            raise ValueError("Format error - invalid 'name'")
        return c

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        struct = vm.StructStackItem(reference_counter)
        struct.append(vm.ByteStringStackItem(self.name))
        array = vm.ArrayStackItem(reference_counter,
                                  list(map(lambda p: p.to_stack_item(reference_counter), self.parameters)))
        struct.append(array)
        return struct


class ContractMethodDescriptor(ContractEventDescriptor, IJson):
    """
    A description of a callable method on a contract.
    """
    def __init__(self, name: str,
                 offset: int,
                 parameters: List[ContractParameterDefinition],
                 return_type: contracts.ContractParameterType,
                 safe: bool):
        """
        Args:
            name: the human readable identifier of the method.
            offset: script offset
            parameters: the list of parameters the method takes.
            return_type: the type of the returned value.
        """
        super(ContractMethodDescriptor, self).__init__(name, parameters)
        self.offset = offset
        self.return_type = return_type
        self.safe = safe

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.name == other.name
                and self.parameters == other.parameters
                and self.offset == other.offset
                and self.return_type == other.return_type
                and self.safe == other.safe)

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = super(ContractMethodDescriptor, self).to_json()
        json.update({
            "returntype": self.return_type.PascalCase(),
            "offset": self.offset,
            "safe": self.safe
        })
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractMethodDescriptor:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValueError: if the manifest name property has an incorrect format.
            ValueError: if the offset is negative.
        """
        c = cls(
            name=contracts.validate_type(json['name'], str),
            offset=contracts.validate_type(json['offset'], int),
            parameters=list(map(lambda p: contracts.ContractParameterDefinition.from_json(p), json['parameters'])),
            return_type=contracts.ContractParameterType[contracts.validate_type(json['returntype'], str).upper()],
            safe=contracts.validate_type(json['safe'], bool)
        )
        if c.name is None or len(c.name) == 0:
            raise ValueError("Format error - invalid 'name'")
        if c.offset < 0:
            raise ValueError("Format error - negative offset not allowed")
        return c

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        struct = cast(vm.StructStackItem, super(ContractMethodDescriptor, self).to_stack_item(reference_counter))
        struct.append(vm.IntegerStackItem(self.return_type.value))
        struct.append(vm.IntegerStackItem(self.offset))
        struct.append(vm.BooleanStackItem(self.safe))
        return struct

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))}> {self.name}"


class ContractABI(IJson):
    """
    The smart contract application binary interface describes the callable events and contracts for a given
    smart contract.
    """
    def __init__(self,
                 methods: List[contracts.ContractMethodDescriptor],
                 events: List[contracts.ContractEventDescriptor]):
        """
        Args:
            byte code.
            methods: the available methods in the contract.
            events: the various events that can be broad casted by the contract.
        """
        self.methods = methods
        self.events = events

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (self.methods == other.methods
                and self.events == other.events)

    def get_method(self, name, parameter_count: int) -> Optional[contracts.ContractMethodDescriptor]:
        """
        Return the ContractMethodDescriptor matching the name (and optional parameter count) or None otherwise.

        Args:
            name: the name of the method to return.
            parameter_count: the expected number of parameters teh method has.
        """
        if parameter_count < -1 or parameter_count > 0xFFFF:
            raise ValueError("Parameter count is out of range")

        if parameter_count >= 0:
            for m in self.methods:
                if m.name == name and len(m.parameters) == parameter_count:
                    return m
            else:
                return None
        else:
            for m in self.methods:
                if m.name == name:
                    return m
            else:
                return None

    def to_json(self) -> dict:
        """
        Convert object into JSON representation.
        """
        json = {
            "methods": list(map(lambda m: m.to_json(), self.methods)),
            "events": list(map(lambda e: e.to_json(), self.events))
        }
        return json

    @classmethod
    def from_json(cls, json: dict) -> ContractABI:
        """
        Parse object out of JSON data.

        Args:
            json: a dictionary.

        Raises:
            KeyError: if the data supplied does not contain the necessary keys.
            ValuError: if the contract has no methods
        """
        c = cls(
            methods=list(map(lambda m: contracts.ContractMethodDescriptor.from_json(m), json['methods'])),
            events=list(map(lambda e: contracts.ContractEventDescriptor.from_json(e), json['events'])),
        )
        if len(c.methods) == 0:
            raise ValueError("Invalid contract - contract has no methods")
        return c

    def to_stack_item(self, reference_counter: vm.ReferenceCounter) -> vm.StackItem:
        struct = vm.StructStackItem(reference_counter)
        struct.append(vm.ArrayStackItem(reference_counter,
                                        list(map(lambda m: m.to_stack_item(reference_counter), self.methods)))
                      )
        struct.append(vm.ArrayStackItem(reference_counter,
                                        list(map(lambda e: e.to_stack_item(reference_counter), self.events)))
                      )
        return struct



import unittest
import aiodns
import logging
import asynctest
import socket
import asyncio
from unittest import mock
from neo3.network import payloads, convenience, node, message, capabilities
from neo3 import network_logger, settings
from datetime import datetime
from neo3.core import msgrouter


class NodeManagerTestCase(asynctest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.nodemgr = convenience.NodeManager()

    def setUp(self) -> None:
        self.nodemgr._reset_for_test()
        node.NeoNode._reset_for_test()
        settings.reset_settings_to_default()

        fake_protocol = object()
        self.node1 = node.NeoNode(fake_protocol)
        self.node2 = node.NeoNode(fake_protocol)

        self.addr1 = payloads.NetworkAddress(address="127.0.0.1:10333")
        self.addr2 = payloads.NetworkAddress(address="127.0.0.2:20333")
        self.node1.address = self.addr1
        self.node2.address = self.addr2

    def test_get_node_addresses(self):
        self.nodemgr.nodes = [self.node1, self.node2]
        addr_list = self.nodemgr.get_node_addresses()
        self.assertEqual([self.addr1, self.addr2], addr_list)

    def test_get_node_by_id(self):
        self.nodemgr.nodes = [self.node1, self.node2]
        n = self.nodemgr.get_node_by_nodeid(self.node1.nodeid)
        self.assertEqual(self.node1, n)

        random_id = 999
        n = self.nodemgr.get_node_by_nodeid(random_id)
        self.assertEqual(None, n)

    def test_get_node_with_height1(self):
        # no connected nodes = early exit
        self.assertEqual(None, self.nodemgr.get_node_with_height(height=1))

    def test_get_node_with_height2(self):
        # can select from 2 , should get the one with the highest weight
        self.node1.best_height = 10
        # increasing error count lowers weight
        self.node1.nodeweight.error_response_count = 5

        self.node2.best_height = 10

        # setup node manager
        self.nodemgr.nodes = [self.node1, self.node2]
        self.assertEqual(self.node2, self.nodemgr.get_node_with_height(height=1))

    def test_get_node_with_height3(self):
        # should return None if none of the nodes have the requested height
        self.node1.best_height = 0
        self.node2.best_height = 0
        # setup node manager
        self.nodemgr.nodes = [self.node1, self.node2]
        self.assertEqual(None, self.nodemgr.get_node_with_height(height=10))

    def test_get_least_failed_node1(self):
        # should return None if none of the nodes have the requested height
        self.node1.best_height = 0
        self.node2.best_height = 0
        # setup node manager
        self.nodemgr.nodes = [self.node1, self.node2]
        ri = convenience.RequestInfo(height=10)
        self.assertEqual(None, self.nodemgr.get_least_failed_node(ri))

    def test_get_least_failed_node2(self):
        # should return the node with the lowest fail count for the request
        # for this test we setup that node1 has failed before, whereas node2 hasn't
        self.node1.best_height = 10
        self.node2.best_height = 10

        ri = convenience.RequestInfo(height=1)
        # pretend node 1 has already failed once for this request
        ri.failed_nodes[self.node1.nodeid] = 1

        # setup node manager
        self.nodemgr.nodes = [self.node1, self.node2]
        self.assertEqual(self.node2, self.nodemgr.get_least_failed_node(ri))

    def test_get_least_failed_node3(self):
        # should not return a node that is in the process of disconnecting but does have the correct height
        self.node1.best_height = 10
        self.node1.disconnecting = True
        self.nodemgr.nodes = [self.node1, self.node2]
        ri = convenience.RequestInfo(height=10)
        self.assertEqual(None, self.nodemgr.get_least_failed_node(ri))

    def test_increase_node_error_count(self):
        self.nodemgr.nodes = [self.node1, self.node2]

        # we expect self.assertLogs to raise an AssertionError because there will be no DEBUG output logs
        # we call self.assertLogs because we want to test for no DBEUG logs being produced
        with self.assertRaises(AssertionError) as assert_log_context:
            with self.assertLogs(network_logger, 'DEBUG') as context:
                    self.nodemgr.increase_node_error_count(self.node1.nodeid)
        self.assertEqual(1, self.node1.nodeweight.error_response_count)
        self.assertEqual(0, len(context.output))

    def test_increase_node_error_count2(self):
        self.nodemgr.nodes = [self.node1, self.node2]
        # set a current value that certainly will exceeds whatever the user configuration will say
        self.node1.nodeweight.error_response_count += 999

        with self.assertLogs(network_logger, 'DEBUG') as context:
            with mock.patch.object(self.node1, 'disconnect'):
                with mock.patch('asyncio.create_task'):
                    self.nodemgr.increase_node_error_count(self.node1.nodeid)
        self.assertEqual(1000, self.node1.nodeweight.error_response_count)
        self.assertIn("max error count threshold exceeded", context.output[0])

    def test_increase_node_timeout_count(self):
        self.nodemgr.nodes = [self.node1, self.node2]

        self.nodemgr.increase_node_timeout_count(self.node1.nodeid)
        self.assertEqual(1, self.node1.nodeweight.timeout_count)

    def test_increase_node_timeout_count2(self):
        # test we detect max count threshold exceeding and schedule a disconnect task
        self.nodemgr.nodes = [self.node1, self.node2]
        # set a current value that certainly will exceeds whatever the user configuration will say
        self.node1.nodeweight.timeout_count += 999

        with self.assertLogs(network_logger, 'DEBUG') as context:
            with mock.patch.object(self.node1, 'disconnect'):
                with mock.patch('asyncio.create_task'):
                    self.nodemgr.increase_node_timeout_count(self.node1.nodeid)
        self.assertEqual(1000, self.node1.nodeweight.timeout_count)
        self.assertIn("max timeout count threshold exceeded", context.output[0])


class AsyncMock(mock.MagicMock):
    async def __call__(self, *args, **kwargs):
        return super(AsyncMock, self).__call__(*args, **kwargs)


class NeoNodeSocketMock(asynctest.SocketMock):
    def __init__(self, loop, hostaddr: str, port: int):
        super(NeoNodeSocketMock, self).__init__()
        self.type = socket.SOCK_STREAM
        self.recv_buffer = bytearray()
        self.loop = loop
        self.hostaddr = hostaddr
        self.port = port
        self.recv_data = self._recv_data()

    def _recv_data(self):
        caps = [capabilities.FullNodeCapability(0)]
        m_send_version = message.Message(msg_type=message.MessageType.VERSION,
                                          payload=payloads.VersionPayload(nonce=123,
                                                                                 user_agent="NEO3-MOCK-CLIENT",
                                                                                 capabilities=caps))
        m_verack = message.Message(msg_type=message.MessageType.VERACK)


        yield m_send_version.to_array()
        yield m_verack.to_array()
        # while True:
        #     yield message.Message(msg_type=message.MessageType.PING,
        #                           payload=payloads.PingPayload(height=0)).to_array()

    def recv(self, max_bytes):
        if not self.recv_buffer:
            try:
                self.recv_buffer.extend(next(self.recv_data))
                asynctest.set_read_ready(self, self.loop)
            except StopIteration:
                # nothing left
                pass

        data = self.recv_buffer[:max_bytes]
        self.recv_buffer = self.recv_buffer[max_bytes:]

        if self.recv_buffer:
            # Some more data to read
            asynctest.set_read_ready(self, self.loop)

        return data

    def send(self, data):
        asynctest.set_read_ready(self, self.loop)
        return len(data)

    def getpeername(self):
        return (self.hostaddr, self.port)


class NodeManagerTestCase2(asynctest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # stdio_handler = logging.StreamHandler()
        # stdio_handler.setLevel(logging.DEBUG)
        # stdio_handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s - %(module)s:%(lineno)s %(message)s"))
        async_logger = logging.getLogger('asyncio')
        # async_logger.addHandler(stdio_handler)
        async_logger.setLevel(logging.DEBUG)

        network_logger = logging.getLogger('neo3.network')
        # network_logger.addHandler(stdio_handler)
        network_logger.setLevel(logging.DEBUG)

        cls.nodemgr = convenience.NodeManager()

    def setUp(self) -> None:
        self.nodemgr._reset_for_test()
        node.NeoNode._reset_for_test()
        settings.reset_settings_to_default()

    async def test_start_shutdown(self):
        settings.register({
            'network': {
                'seedlist': ['127.0.0.1:1111'],
                'magic': 769
            }
        })

        self.nodemgr.start()

        def client_provider():
            addresses = [('127.0.0.1', 1111)]
            for pair in addresses:
                yield NeoNodeSocketMock(self.loop, pair[0], pair[1])

        # configure nodemanager to use a mock client
        self.nodemgr._test_client_provider = client_provider

        call_back_result = None
        def client_connect_done(node_instance, failure):
            nonlocal call_back_result
            call_back_result = node_instance

        # listen to a connection done event
        msgrouter.on_client_connect_done += client_connect_done

        # initial state
        self.assertFalse(self.nodemgr.is_running)
        # advance loop
        await asyncio.sleep(0.1)
        self.assertTrue(self.nodemgr.is_running)
        # 4 looping service tasks
        self.assertEqual(4, len(self.nodemgr.tasks))
        self.assertIsInstance(call_back_result, node.NeoNode)
        await self.nodemgr.shutdown()
        for t in self.nodemgr.tasks:
            self.assertTrue(t.done())

    async def test_fill_open_connection_spots_no_clients(self):
        # create 1 open spot
        self.nodemgr.max_clients = 1
        # ensure we have an address that can be connected to
        node.NeoNode.addresses = [payloads.NetworkAddress(address='127.0.0.1:1111')]

        with self.assertLogs(network_logger, 'DEBUG') as context:
            with mock.patch('asyncio.create_task'):
                await self.nodemgr._fill_open_connection_spots()
        self.assertIn("Found 1 open pool spots", context.output[0])
        self.assertIn("Adding 127.0.0.1:1111 to connection queue", context.output[1])

    async def test_fill_open_connection_spots_dont_queue_addr_that_is_already_queued(self):
        fake_protocol = object()
        self.node1 = node.NeoNode(fake_protocol)

        # ensure we have an address that can be connected to
        addr = payloads.NetworkAddress(address='127.0.0.1:1111')
        # create 1 open spot
        self.nodemgr.max_clients = 2
        node.NeoNode.addresses = [addr]
        # and ensure we pretend we have already queued this address
        self.nodemgr.queued_addresses = [addr]

        with self.assertLogs(network_logger, 'DEBUG') as context:
            with mock.patch('asyncio.create_task'):
                await self.nodemgr._fill_open_connection_spots()
        self.assertIn("Found 1 open pool spots", context.output[0])
        # context.output length should only contain the above debug message, nothing more
        self.assertEqual(1, len(context.output))

    async def test_fill_open_connection_spots_no_addresses_to_fill_spots(self):
        # we first test that no errors occur if we have open spots, but we still meet our minimum required clients
        # this should not cause any error count increases
        self.nodemgr.min_clients = 1
        self.nodemgr.max_clients = 2
        # just a place holder to have a count matching min_clients
        self.nodemgr.nodes = [object()]
        node.NeoNode.addresses = []

        with self.assertLogs(network_logger, 'DEBUG') as context:
            await self.nodemgr._fill_open_connection_spots()
        self.assertIn("Found 1 open pool spots, trying to add nodes...", context.output[0])
        self.assertIn("No addresses available to fill spots. However, minimum clients still satisfied", context.output[1])

        # next we clear the nodes and start inducing errors for not being able to fill the open spots
        self.nodemgr.nodes = []
        with self.assertLogs(network_logger, 'DEBUG') as context:
            await self.nodemgr._fill_open_connection_spots()
        self.assertIn("Found 2 open pool spots, trying to add nodes...", context.output[0])
        self.assertIn("Increasing pool spot error count to 1", context.output[1])

        with self.assertLogs(network_logger, 'DEBUG') as context:
            await self.nodemgr._fill_open_connection_spots()
        self.assertIn("Found 2 open pool spots, trying to add nodes...", context.output[0])
        self.assertIn("Increasing pool spot error count to 2", context.output[1])

        # 3rd time we reached our threshold
        # let's also setup an address in POOR state that should be reset to NEW
        addr = payloads.NetworkAddress(address='127.0.0.1:1111')
        addr.set_state_poor()
        node.NeoNode.addresses = [addr]
        with self.assertLogs(network_logger, 'DEBUG') as context:
            await self.nodemgr._fill_open_connection_spots()
        self.assertIn("Found 2 open pool spots, trying to add nodes...", context.output[0])
        self.assertIn("Recycling old addresses", context.output[1])
        self.assertEqual(0, self.nodemgr.MAX_NODE_POOL_ERROR_COUNT)
        self.assertTrue(node.NeoNode.addresses[0].is_state_new)

    async def test_fill_open_connection_spots_node_timeout_error(self):
        self.nodemgr.min_clients = 1
        self.nodemgr.max_clients = 2
        addr = payloads.NetworkAddress(address='127.0.0.1:1111')
        node.NeoNode.addresses = [addr]

        call_back_result = None
        def client_connect_done(node_instance, failure):
            nonlocal call_back_result
            call_back_result = failure

        # listen to a connection done event
        msgrouter.on_client_connect_done += client_connect_done

        with self.assertLogs(network_logger, 'DEBUG') as context:
            self.loop.create_connection = AsyncMock()
            self.loop.create_connection.side_effect = asyncio.TimeoutError()
            await self.nodemgr._fill_open_connection_spots()

        # advance loop
        await asyncio.sleep(0.1)
        self.assertEqual("Timed out", call_back_result[1])
        self.assertIn("Found 2 open pool spots, trying to add nodes...", context.output[0])


class NodeManagerTimedTestCase(asynctest.ClockedTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.nodemgr = convenience.NodeManager()

    def setUp(self) -> None:
        self.nodemgr._reset_for_test()
        node.NeoNode._reset_for_test()
        settings.reset_settings_to_default()

        fake_protocol = object()
        self.node1 = node.NeoNode(fake_protocol)
        self.node2 = node.NeoNode(fake_protocol)

        self.addr1 = payloads.NetworkAddress(address="127.0.0.1:10333")
        self.addr2 = payloads.NetworkAddress(address="127.0.0.2:20333")
        self.node1.address = self.addr1
        self.node2.address = self.addr2

    @asynctest.SkipTest
    async def test_utility_run_in_loop(self):
        counter = 0
        async def coro():
            nonlocal counter
            counter += 1

        self.nodemgr._run_in_loop(coro, 1)
        # trigger the task by ticking the loop once
        await self.advance(0)
        self.assertEqual(1, counter)
        await self.advance(0.9)
        self.assertEqual(1, counter)
        await self.advance(0.1)
        self.assertEqual(2, counter)
        self.nodemgr.tasks[0].cancel()

    async def test_monitor_node_height_disconnect_when_treshold_exceeded(self):
        # ensure we exceed the threshold
        self.node1.best_height_last_update = datetime.utcnow().timestamp() - self.nodemgr.MAX_HEIGHT_UPDATE_DURATION - 1
        self.nodemgr.nodes = [self.node1]
        with self.assertLogs(network_logger, 'DEBUG') as context:
            with mock.patch.object(self.node1, 'disconnect') as patched_node_disconnect:
                with mock.patch('asyncio.create_task'):
                    await self.nodemgr._monitor_node_height()
        self.assertIn("max height update threshold exceeded", context.output[0])
        patched_node_disconnect.assert_called_with(reason=payloads.DisconnectReason.POOR_PERFORMANCE)

    async def test_monitor_node_height_within_limits(self):
        # test that if the last height update timestamp of a node is within the treshold
        # that we only ask for another update
        settings.storage.use_default = False
        self.nodemgr.nodes = [self.node1]
        with self.assertLogs(network_logger, 'DEBUG') as context:
            with mock.patch.object(self.node1, 'send_message', side_effect=AsyncMock()) as patched_node_send_msg:
                await self.nodemgr._monitor_node_height()
        self.assertIn(" to send us a height update (PING)", context.output[0])
        self.assertGreater(len(patched_node_send_msg.call_args), 0)
        msg = patched_node_send_msg.call_args[0][0]
        self.assertEqual(message.MessageType.PING, msg.type)
        # finally test that our tasks is created
        self.assertEqual(1, len(self.nodemgr.tasks))
        await self.advance(1)
        # and the call back called once done.
        self.assertEqual(0, len(self.nodemgr.tasks))

    async def test_query_addresses(self):
        self.nodemgr.nodes = [self.node1]
        with self.assertLogs(network_logger, 'DEBUG') as context:
            with mock.patch.object(self.node1, 'request_address_list', side_effect=AsyncMock()) as patched_node_req_addr:
                await self.nodemgr._query_addresses()
        self.assertIn("Asking node ", context.output[1])
        self.assertIn("for its address list", context.output[1])
        patched_node_req_addr.assert_called_once()
        # finally test that our tasks is created
        self.assertEqual(1, len(self.nodemgr.tasks))
        await self.advance(1)
        # and the call back called once done.
        self.assertEqual(0, len(self.nodemgr.tasks))

    async def test_seed_list_processing(self):
        # specific testing of named addresses
        settings.network.seedlist = ['google.com:1111']

        # for whatever reason I can't find the right way to mock the property, tried PropertyMock and what not.
        # this works *shrug*
        class wtf:
            host = '127.0.0.123'
        result = [wtf()]
        with mock.patch('aiodns.DNSResolver.query', side_effect=asynctest.CoroutineMock(return_value=result)) as query_result:
            await self.nodemgr._process_seed_list_addresses()
        self.assertEqual(1, len(node.NeoNode.addresses))

        with self.assertLogs(network_logger, 'DEBUG') as context:
            with mock.patch('aiodns.DNSResolver.query', side_effect=aiodns.error.DNSError) as query_result:
                await self.nodemgr._process_seed_list_addresses()
        self.assertIn("Skipping google.com, address could not be resolved", context.output[0])
